from datetime import datetime, timedelta
from typing import TYPE_CHECKING

from dolphie.DataTypes import ConnectionSource, ConnectionStatus
from dolphie.Modules.PerformanceSchemaMetrics import PerformanceSchemaMetrics
from dolphie.Modules.Queries import MySQLQueries, ProxySQLQueries
from dolphie.Panels import MetadataLocks as MetadataLocksPanel
from dolphie.Panels import Processlist as ProcesslistPanel
from dolphie.Panels import ProxySQLProcesslist as ProxySQLProcesslistPanel
from dolphie.Panels import Replication as ReplicationPanel
from loguru import logger

if TYPE_CHECKING:
    from dolphie.App import DolphieApp
    from dolphie.Modules.TabManager import Tab


class WorkerDataProcessor:
    """Manages polling data processing and screen refresh operations for worker threads.

    This class encapsulates all data processing logic, screen refresh operations,
    and monitoring of various system changes.
    """

    def __init__(self, app: "DolphieApp"):
        """Initialize the WorkerDataProcessor.

        Args:
            app: Reference to the main DolphieApp instance
        """
        self.app = app

    def process_mysql_data(self, tab: "Tab"):
        """Process MySQL data for a given tab."""
        dolphie = tab.dolphie

        global_variables = dolphie.main_db_connection.fetch_status_and_variables("variables")
        self.monitor_global_variable_change(tab=tab, old_data=dolphie.global_variables, new_data=global_variables)
        dolphie.global_variables = global_variables

        # At this point, we're connected so we need to do a few things
        if dolphie.connection_status == ConnectionStatus.connecting:
            # Called from worker thread, use call_from_thread
            self.app.call_from_thread(
                self.app.tab_manager.update_connection_status, tab=tab, connection_status=ConnectionStatus.connected
            )
            dolphie.host_version = dolphie.parse_server_version(dolphie.global_variables.get("version"))
            dolphie.get_group_replication_metadata()
            dolphie.configure_mysql_variables()
            dolphie.validate_metadata_locks_enabled()

        global_status = dolphie.main_db_connection.fetch_status_and_variables("status")
        self.monitor_uptime_change(
            tab=tab, old_uptime=dolphie.global_status.get("Uptime", 0), new_uptime=global_status.get("Uptime", 0)
        )
        dolphie.global_status = global_status
        # If the server doesn't support Innodb_lsn_current, use Innodb_os_log_written instead
        # which has less precision, but it's good enough. Used for calculating the percentage of redo log used
        if not dolphie.global_status.get("Innodb_lsn_current"):
            dolphie.global_status["Innodb_lsn_current"] = dolphie.global_status.get("Innodb_os_log_written")

        dolphie.innodb_metrics = dolphie.main_db_connection.fetch_status_and_variables("innodb_metrics")

        if dolphie.galera_cluster and dolphie.panels.replication.visible:
            dolphie.main_db_connection.execute(MySQLQueries.get_galera_cluster_members)
            dolphie.galera_cluster_members = dolphie.main_db_connection.fetchall()

        dolphie.replication_status = ReplicationPanel.fetch_replication_data(tab)

        # Manage our replicas — use processlist for discovery (real connection IP)
        # and SHOW REPLICAS/SHOW SLAVE HOSTS for port correlation
        if dolphie.connection_source_alt == ConnectionSource.mariadb:
            find_replicas_query = (
                MySQLQueries.mariadb_find_replicas
                if dolphie.performance_schema_enabled
                else MySQLQueries.pl_find_replicas
            )
        elif dolphie.performance_schema_enabled and dolphie.is_mysql_version_at_least("5.7"):
            find_replicas_query = MySQLQueries.ps_find_replicas
        else:
            find_replicas_query = MySQLQueries.pl_find_replicas

        dolphie.main_db_connection.execute(find_replicas_query)
        available_replicas = dolphie.main_db_connection.fetchall()

        if not dolphie.daemon_mode:
            # Refresh port data when the number of replicas changes
            if len(available_replicas) != len(dolphie.replica_manager.available_replicas):
                use_show_replicas = (
                    dolphie.connection_source_alt != ConnectionSource.mariadb
                    and dolphie.is_mysql_version_at_least("8.0.22")
                )
                query = MySQLQueries.show_replicas if use_show_replicas else MySQLQueries.show_slave_hosts

                dolphie.main_db_connection.execute(query)
                ports_replica_data = dolphie.main_db_connection.fetchall()

                dolphie.replica_manager.ports = {}
                if dolphie.connection_source_alt == ConnectionSource.mariadb:
                    for row in ports_replica_data:
                        dolphie.replica_manager.ports[row.get("Server_id")] = {
                            "port": row.get("Port"),
                            "host": row.get("Host"),
                            "in_use": False,
                        }
                else:
                    uuid_key = "Replica_UUID" if use_show_replicas else "Slave_UUID"
                    for row in ports_replica_data:
                        dolphie.replica_manager.ports[row.get(uuid_key)] = {
                            "port": row.get("Port"),
                            "host": row.get("Host"),
                        }

        dolphie.replica_manager.available_replicas = available_replicas

        if dolphie.is_mysql_version_at_least("8.2.0") and dolphie.connection_source_alt != ConnectionSource.mariadb:
            dolphie.main_db_connection.execute(MySQLQueries.show_binary_log_status)
        else:
            dolphie.main_db_connection.execute(MySQLQueries.show_master_status)

        previous_position = dolphie.binlog_status.get("Position")
        dolphie.binlog_status = dolphie.main_db_connection.fetchone()

        if previous_position is None:
            dolphie.binlog_status["Diff_Position"] = 0
        elif previous_position > dolphie.binlog_status["Position"]:
            dolphie.binlog_status["Diff_Position"] = "Binlog Rotated"
        else:
            dolphie.binlog_status["Diff_Position"] = dolphie.binlog_status["Position"] - previous_position

        if dolphie.panels.processlist.visible:
            dolphie.processlist_threads = ProcesslistPanel.fetch_data(tab)

        if dolphie.panels.replication.visible and (dolphie.innodb_cluster or dolphie.innodb_cluster_read_replica):
            dolphie.main_db_connection.execute(MySQLQueries.get_clustersets)
            dolphie.innodb_cluster_clustersets = dolphie.main_db_connection.fetchall()

        if dolphie.performance_schema_enabled:
            dolphie.main_db_connection.execute(MySQLQueries.ps_disk_io)
            dolphie.disk_io_metrics = dolphie.main_db_connection.fetchone()

            # MariaDB uses slave_parallel_threads; MySQL uses replica_parallel_workers
            if dolphie.connection_source_alt == ConnectionSource.mariadb:
                parallel_workers = dolphie.global_variables.get("slave_parallel_threads", 0)
            else:
                parallel_workers = dolphie.global_variables.get("replica_parallel_workers", 0)

            if dolphie.connection_source_alt == ConnectionSource.mariadb:
                has_applier_status = dolphie.is_mysql_version_at_least("10.5")
            else:
                has_applier_status = dolphie.is_mysql_version_at_least("8.0")

            if (
                has_applier_status
                and dolphie.replication_status
                and dolphie.panels.replication.visible
                and parallel_workers > 1
            ):
                dolphie.main_db_connection.execute(MySQLQueries.replication_applier_status)
                dolphie.replication_applier_status["data"] = dolphie.main_db_connection.fetchall()

                # Calculate the difference in total_thread_events for each worker + all workers
                for row in dolphie.replication_applier_status["data"]:
                    total_thread_events = row["total_thread_events"]
                    thread_id = row.get("thread_id")

                    # Handle the ROLLUP row, which contains the total for all threads
                    if not thread_id:
                        dolphie.replication_applier_status["diff_all"] = (
                            total_thread_events
                            - dolphie.replication_applier_status.get("previous_all", total_thread_events)
                        )
                        dolphie.replication_applier_status["previous_all"] = total_thread_events
                        continue

                    dolphie.replication_applier_status[f"diff_{thread_id}"] = (
                        total_thread_events
                        - dolphie.replication_applier_status.get(f"previous_{thread_id}", total_thread_events)
                    )
                    dolphie.replication_applier_status[f"previous_{thread_id}"] = total_thread_events

            if (
                not dolphie.daemon_mode
                and dolphie.is_mysql_version_at_least("8.0.30")
                and dolphie.connection_source_alt != ConnectionSource.mariadb
            ):
                active_redo_logs_count = dolphie.main_db_connection.fetch_value_from_field(
                    MySQLQueries.active_redo_logs, "count"
                )
                dolphie.global_status["Active_redo_log_count"] = active_redo_logs_count

            if dolphie.group_replication or dolphie.innodb_cluster:
                if dolphie.is_mysql_version_at_least("8.0.13"):
                    dolphie.group_replication_data["write_concurrency"] = (
                        dolphie.main_db_connection.fetch_value_from_field(
                            MySQLQueries.group_replication_get_write_concurrency, "write_concurrency"
                        )
                    )

                dolphie.main_db_connection.execute(MySQLQueries.get_group_replication_members)
                dolphie.group_replication_members = dolphie.main_db_connection.fetchall()

            if dolphie.is_mysql_version_at_least("5.7"):
                dolphie.metadata_locks = {}
                if dolphie.metadata_locks_enabled and dolphie.panels.metadata_locks.visible:
                    dolphie.metadata_locks = MetadataLocksPanel.fetch_data(tab)

                if dolphie.panels.ddl.visible:
                    dolphie.main_db_connection.execute(MySQLQueries.ddls)
                    dolphie.ddl = dolphie.main_db_connection.fetchall()

                if dolphie.panels.pfs_metrics.visible:
                    # Reset the PFS metrics deltas if we're in daemon mode and it's been 10 minutes since the last reset
                    # This is to keep a realistic point-in-time view of the metrics
                    time_since_reset = datetime.now().astimezone() - dolphie.pfs_metrics_last_reset_time
                    if dolphie.daemon_mode and time_since_reset >= timedelta(minutes=10):
                        dolphie.reset_pfs_metrics_deltas()

                    dolphie.main_db_connection.execute(MySQLQueries.file_summary_by_instance)
                    file_io_data = dolphie.main_db_connection.fetchall()
                    if not dolphie.file_io_data:
                        dolphie.file_io_data = PerformanceSchemaMetrics(file_io_data, "file_io", "FILE_NAME")
                    else:
                        dolphie.file_io_data.update_internal_data(file_io_data)

                    dolphie.main_db_connection.execute(MySQLQueries.table_io_waits_summary_by_table)
                    table_io_waits_data = dolphie.main_db_connection.fetchall()
                    if not dolphie.table_io_waits_data:
                        dolphie.table_io_waits_data = PerformanceSchemaMetrics(
                            table_io_waits_data, "table_io", "OBJECT_TABLE"
                        )
                    else:
                        dolphie.table_io_waits_data.update_internal_data(table_io_waits_data)

                if dolphie.panels.statements_summary.visible:
                    if (
                        dolphie.is_mysql_version_at_least("8.0")
                        and dolphie.connection_source_alt != ConnectionSource.mariadb
                    ):
                        dolphie.main_db_connection.execute(MySQLQueries.table_statements_summary_by_digest_80)
                    else:
                        dolphie.main_db_connection.execute(MySQLQueries.table_statements_summary_by_digest)

                    statements_summary_data = dolphie.main_db_connection.fetchall()
                    if not dolphie.statements_summary_data:
                        dolphie.statements_summary_data = PerformanceSchemaMetrics(
                            statements_summary_data, "statements_summary", "digest"
                        )
                    else:
                        dolphie.statements_summary_data.update_internal_data(statements_summary_data)

    def process_proxysql_data(self, tab: "Tab"):
        """Process ProxySQL data for a given tab."""
        dolphie = tab.dolphie

        global_variables = dolphie.main_db_connection.fetch_status_and_variables("variables")
        self.monitor_global_variable_change(tab=tab, old_data=dolphie.global_variables, new_data=global_variables)
        dolphie.global_variables = global_variables

        if dolphie.connection_status == ConnectionStatus.connecting:
            # Called from worker thread, use call_from_thread
            self.app.call_from_thread(
                self.app.tab_manager.update_connection_status, tab=tab, connection_status=ConnectionStatus.connected
            )
            dolphie.host_version = dolphie.parse_server_version(dolphie.global_variables.get("admin-version"))

        global_status = dolphie.main_db_connection.fetch_status_and_variables("mysql_stats")
        self.monitor_uptime_change(
            tab=tab,
            old_uptime=dolphie.global_status.get("ProxySQL_Uptime", 0),
            new_uptime=global_status.get("ProxySQL_Uptime", 0),
        )
        dolphie.global_status = global_status

        dolphie.main_db_connection.execute(ProxySQLQueries.command_stats)
        dolphie.proxysql_command_stats = dolphie.main_db_connection.fetchall()

        # Here, we're going to format the command stats to match the global status keys of
        # MySQL and get total count of queries
        total_queries_count = 0
        query_types_for_total = ["SELECT", "INSERT", "UPDATE", "DELETE", "REPLACE", "SET", "CALL"]
        for row in dolphie.proxysql_command_stats:
            total_cnt = 0
            if row["Command"] in query_types_for_total:
                total_cnt = int(row["Total_cnt"])
                total_queries_count += total_cnt

            dolphie.global_status[f"Com_{row['Command'].lower()}"] = total_cnt

        # Add the total queries to the global status
        dolphie.global_status["Queries"] = total_queries_count

        dolphie.main_db_connection.execute(ProxySQLQueries.connection_pool_data)
        data = dolphie.main_db_connection.fetchone()

        if dolphie.global_status.get("Client_Connections_connected", 0):
            dolphie.global_status["proxysql_multiplex_efficiency_ratio"] = round(
                100
                - (
                    (
                        int(data.get("connection_pool_connections", 0))
                        / dolphie.global_status.get("Client_Connections_connected", 0)
                    )
                    * 100
                ),
                2,
            )
        else:
            dolphie.global_status["proxysql_multiplex_efficiency_ratio"] = 100

        if dolphie.panels.proxysql_hostgroup_summary.visible:
            dolphie.main_db_connection.execute(ProxySQLQueries.hostgroup_summary)

            previous_values = {}
            columns_to_calculate_per_sec = ["Queries", "Bytes_data_sent", "Bytes_data_recv"]

            # Store previous values for each row
            for row in dolphie.proxysql_hostgroup_summary:
                row_id = f"{row['hostgroup']}_{row['srv_host']}_{row['srv_port']}"

                for column_key in columns_to_calculate_per_sec:
                    previous_values.setdefault(row_id, {})[column_key] = int(row.get(column_key, 0))

            # Fetch the updated hostgroup summary
            dolphie.proxysql_hostgroup_summary = dolphie.main_db_connection.fetchall()

            # Calculate the values per second
            for row in dolphie.proxysql_hostgroup_summary:
                row_id = f"{row['hostgroup']}_{row['srv_host']}_{row['srv_port']}"

                if row_id in previous_values:  # Ensure we have previous values for this row_id
                    for column_key in columns_to_calculate_per_sec:
                        previous_value = previous_values[row_id].get(column_key, 0)
                        current_value = int(row.get(column_key, 0))

                        value_per_sec = (
                            (current_value - previous_value) / dolphie.polling_latency
                            if dolphie.polling_latency > 0
                            else 0
                        )
                        row[f"{column_key}_per_sec"] = round(value_per_sec)

        if dolphie.panels.processlist.visible:
            dolphie.processlist_threads = ProxySQLProcesslistPanel.fetch_data(tab)

        if dolphie.panels.proxysql_mysql_query_rules.visible:
            dolphie.main_db_connection.execute(ProxySQLQueries.query_rules_summary)
            dolphie.proxysql_mysql_query_rules = dolphie.main_db_connection.fetchall()

    def refresh_screen_proxysql(self, tab: "Tab"):
        """Refresh the ProxySQL screen for a given tab."""
        dolphie = tab.dolphie

        if tab.loading_indicator.display:
            tab.loading_indicator.display = False

        # Loop each panel and refresh it
        for panel in dolphie.panels.get_all_panels():
            if panel.visible:
                # Skip the graphs panel since it's handled separately
                if panel.name == dolphie.panels.graphs.name:
                    continue

                self.app.refresh_panel(tab, panel.name)

                if panel.name == dolphie.panels.dashboard.name and dolphie.metric_manager.metrics.dml.Queries.values:
                    # Update the sparkline for queries per second
                    tab.sparkline.data = dolphie.metric_manager.metrics.dml.Queries.values
                    tab.sparkline.refresh()

        # Refresh the graph(s) for the selected tab
        self.app.update_graphs(tab.metric_graph_tabs.get_pane(tab.metric_graph_tabs.active).name)

        tab.refresh_replay_dashboard_section()

        # We take a snapshot of the processlist to be used for commands
        # since the data can change after a key is pressed
        if not dolphie.daemon_mode:
            dolphie.processlist_threads_snapshot = dolphie.processlist_threads.copy()

    def refresh_screen_mysql(self, tab: "Tab"):
        """Refresh the MySQL screen for a given tab."""
        dolphie = tab.dolphie

        if tab.loading_indicator.display:
            tab.loading_indicator.display = False

        # Loop each panel and refresh it
        for panel in dolphie.panels.get_all_panels():
            if panel.visible:
                # Skip the graphs panel since it's handled separately
                if panel.name == dolphie.panels.graphs.name:
                    continue

                self.app.refresh_panel(tab, panel.name)

                if panel.name == dolphie.panels.dashboard.name and dolphie.metric_manager.metrics.dml.Queries.values:
                    # Update the sparkline for queries per second
                    tab.sparkline.data = dolphie.metric_manager.metrics.dml.Queries.values
                    tab.sparkline.refresh()

        # Refresh the graph(s) for the selected tab
        self.app.update_graphs(tab.metric_graph_tabs.get_pane(tab.metric_graph_tabs.active).name)

        tab.refresh_replay_dashboard_section()

        # We take a snapshot of the processlist to be used for commands
        # since the data can change after a key is pressed
        if not dolphie.daemon_mode:
            dolphie.processlist_threads_snapshot = dolphie.processlist_threads.copy()

    def monitor_global_variable_change(self, tab: "Tab", old_data: dict, new_data: dict):
        """Monitor and notify about global variable changes."""
        if not old_data:
            return

        dolphie = tab.dolphie

        # gtid is always changing so we don't want to alert on that
        # The others are ones I've found to be spammy due to monitoring tools changing them
        exclude_variables = {"gtid", "innodb_thread_sleep_delay"}

        # Add to exclude_variables with user specified variables
        if dolphie.exclude_notify_global_vars:
            exclude_variables.update(v.lower() for v in dolphie.exclude_notify_global_vars)

        for variable, new_value in new_data.items():
            if any(item in variable for item in exclude_variables):
                continue

            old_value = old_data.get(variable)
            if old_value != new_value:
                tab.replay_manager.capture_global_variable_change(variable, old_value, new_value)

                # read_only notification/log message is handled by monitor_read_only_change()
                if variable == "read_only":
                    continue

                logger.info(f"Global variable {variable} changed: {old_value} -> {new_value}")

                # Skip UI notifications in daemon mode since the TUI is headless
                if dolphie.daemon_mode:
                    continue

                # If the tab is not active, include the host in the notification
                include_host = ""
                if self.app.tab_manager.active_tab.id != tab.id:
                    include_host = f"Host:      [$light_blue]{dolphie.host_with_port}[/$light_blue]\n"
                self.app.notify(
                    f"[b][$dark_yellow]{variable}[/b][/$dark_yellow]\n"
                    f"{include_host}"
                    f"Old Value: [$highlight]{old_value}[/$highlight]\n"
                    f"New Value: [$highlight]{new_value}[/$highlight]",
                    title="Global Variable Change",
                    severity="warning",
                    timeout=15,
                )

    def monitor_uptime_change(self, tab: "Tab", old_uptime: int, new_uptime: int):
        """Monitor and handle uptime changes (e.g., server restarts)."""
        if old_uptime > new_uptime:
            formatted_old_uptime = str(timedelta(seconds=old_uptime))
            formatted_new_uptime = str(timedelta(seconds=new_uptime))

            tab.replay_manager.capture_global_variable_change("Uptime", formatted_old_uptime, formatted_new_uptime)

            logger.info(f"Uptime changed: {formatted_old_uptime} -> {formatted_new_uptime}")

            # Reset data for Performance Schema metrics since those tables are reset on server restart
            tab.dolphie.reset_pfs_metrics_deltas(reset_fully=True)

    def monitor_read_only_change(self, tab: "Tab"):
        """Monitor and notify about read-only status changes."""
        dolphie = tab.dolphie

        if dolphie.connection_source == ConnectionSource.proxysql:
            return

        current_ro_status = dolphie.global_variables.get("read_only")
        formatted_ro_status = ConnectionStatus.read_only if current_ro_status == "ON" else ConnectionStatus.read_write

        if (
            dolphie.connection_status in (ConnectionStatus.read_write, ConnectionStatus.read_only)
            and dolphie.connection_status != formatted_ro_status
        ):
            status = "read-only" if current_ro_status == "ON" else "read/write"
            message = (
                f"Host [$light_blue]{dolphie.host_with_port}[/$light_blue]"
                f" is now [$b_highlight]{status}[/$b_highlight]"
            )

            # Warn if a primary/standalone is unexpectedly read-only
            if current_ro_status == "ON" and (
                (not dolphie.replication_status and not dolphie.group_replication)
                or (dolphie.group_replication and dolphie.is_group_replication_primary)
            ):
                message += " ([$dark_yellow]SHOULD BE READ/WRITE?[/$dark_yellow])"

            logger.warning(f"Read-only mode changed: {dolphie.connection_status} -> {formatted_ro_status}")
            self.app.notify(title="Read-only mode change", message=message, severity="warning", timeout=15)

            self.app.tab_manager.update_connection_status(tab=tab, connection_status=formatted_ro_status)
        elif dolphie.connection_status == ConnectionStatus.connected:
            self.app.tab_manager.update_connection_status(tab=tab, connection_status=formatted_ro_status)



import csv
import re
import threading
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING

from dolphie.DataTypes import (
    ConnectionSource,
    HotkeyCommands,
    ProcesslistThread,
    ProxySQLProcesslistThread,
)
from dolphie.Modules.Functions import (
    escape_markup,
    format_bytes,
    format_number,
    format_query,
)
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.Queries import MySQLQueries, ProxySQLQueries
from dolphie.Widgets.CommandModal import CommandModal
from dolphie.Widgets.CommandScreen import CommandScreen
from dolphie.Widgets.EventLogScreen import EventLog
from dolphie.Widgets.ProxySQLThreadScreen import ProxySQLThreadScreen
from dolphie.Widgets.ThreadScreen import ThreadScreen
from rich import box
from rich.align import Align
from rich.console import Group
from rich.style import Style
from rich.table import Table
from sqlparse import format as sqlformat
from textual.widgets import Button

if TYPE_CHECKING:
    from dolphie.App import DolphieApp


class KeyEventManager:
    """This module manages all keyboard event processing.

    This includes both immediate key event handling and background command execution
    in threads.
    """

    def __init__(self, app: "DolphieApp"):
        """Initialize the KeyEventManager.

        Args:
            app: Reference to the main DolphieApp instance
        """
        self.app = app

        # Debouncing to prevent rapid key presses from overwhelming the system
        self.last_key_time = {}
        self.default_debounce_interval = timedelta(milliseconds=50)

        # Custom debounce intervals for specific keys that trigger expensive operations
        self.key_debounce_intervals = {
            "left_square_bracket": timedelta(milliseconds=100),  # Replay backward
            "right_square_bracket": timedelta(milliseconds=100),  # Replay forward
            "space": timedelta(milliseconds=300),  # Start worker
            "minus": timedelta(milliseconds=300),  # Remove tab (destructive)
        }

    async def process_key_event(self, key: str) -> None:
        """Process a keyboard event and execute the corresponding action.

        This method handles all keyboard shortcuts and commands in the application,
        from panel switching to data filtering to command execution.

        Args:
            key: The key that was pressed
        """
        tab = self.app.tab_manager.active_tab
        if not tab:
            return

        # Apply debouncing to prevent rapid key presses
        now = datetime.now().astimezone()
        debounce_interval = self.key_debounce_intervals.get(key, self.default_debounce_interval)
        last_time = self.last_key_time.get(key, datetime.min.replace(tzinfo=timezone.utc))

        if now - last_time < debounce_interval:
            return  # Key press is too soon, ignore it

        self.last_key_time[key] = now

        screen_data = None
        dolphie = tab.dolphie

        # Validate key is a valid command (excluding special keys)
        if key not in self.app.command_manager.exclude_keys:
            if not self.app.command_manager.get_commands(dolphie.replay_file, dolphie.connection_source).get(key):
                self.app.notify(
                    f"Key [$highlight]{key}[/$highlight] is not a valid command",
                    severity="warning",
                )
                return

            # Prevent commands from being run if the secondary connection is processing a query already
            if dolphie.secondary_db_connection and dolphie.secondary_db_connection.is_running_query:
                self.app.notify("There's already a command running - please wait for it to finish")
                return

            if not dolphie.main_db_connection.is_connected() and not dolphie.replay_file:
                self.app.notify("You must be connected to a host to use commands")
                return

        if self.app.tab_manager.loading_hostgroups:
            self.app.notify("You can't run commands while hosts are connecting as a hostgroup")
            return

        # Panel switching commands (1-8)
        if key == "1":
            self.app.toggle_panel(dolphie.panels.dashboard.name)

        elif key == "2":
            self.app.tab_manager.active_tab.processlist_datatable.clear()
            self.app.toggle_panel(dolphie.panels.processlist.name)

        elif key == "3":
            self.app.toggle_panel(dolphie.panels.graphs.name)
            self.app.update_graphs(tab.metric_graph_tabs.get_pane(tab.metric_graph_tabs.active).name)

        elif key == "4":
            if dolphie.connection_source == ConnectionSource.proxysql:
                self.app.toggle_panel(dolphie.panels.proxysql_hostgroup_summary.name)
                dolphie.proxysql_per_second_data.clear()
                self.app.tab_manager.active_tab.proxysql_hostgroup_summary_datatable.clear()
                return

            has_replication_data = any(
                [
                    dolphie.replication_status,
                    dolphie.galera_cluster,
                    dolphie.group_replication,
                    dolphie.innodb_cluster,
                    dolphie.innodb_cluster_read_replica,
                    # Replicas can only be displayed in live mode since we connect to each one
                    not dolphie.replay_file and dolphie.replica_manager.available_replicas,
                ]
            )
            if not has_replication_data:
                self.app.notify("Replication panel has no data to display")
                return

            self.app.toggle_panel(dolphie.panels.replication.name)
            tab.toggle_entities_displays()

            if dolphie.panels.replication.visible:
                if dolphie.replica_manager.available_replicas:
                    # No loading animation necessary for replay mode
                    if not dolphie.replay_file:
                        tab.replicas_loading_indicator.display = True
                        tab.replicas_title.update(
                            f"[$white][b]Loading [$highlight]{len(dolphie.replica_manager.available_replicas)}"
                            "[/$highlight] replicas...\n"
                        )

                tab.toggle_replication_panel_components()
            else:
                tab.remove_replication_panel_components()

        elif key == "5":
            if dolphie.connection_source == ConnectionSource.proxysql:
                self.app.toggle_panel(dolphie.panels.proxysql_mysql_query_rules.name)
                return

            if not dolphie.metadata_locks_enabled and not dolphie.replay_file:
                self.app.notify(
                    "Metadata Locks panel requires MySQL 5.7+ with Performance Schema enabled along with "
                    "[$highlight]wait/lock/metadata/sql/mdl[/$highlight] enabled in setup_instruments table"
                )
                return

            self.app.toggle_panel(dolphie.panels.metadata_locks.name)
            self.app.tab_manager.active_tab.metadata_locks_datatable.clear()

        elif key == "6":
            if dolphie.connection_source == ConnectionSource.proxysql:
                self.app.toggle_panel(dolphie.panels.proxysql_command_stats.name)
            else:
                if not dolphie.is_mysql_version_at_least("5.7") or not dolphie.performance_schema_enabled:
                    self.app.notify("DDL panel requires MySQL 5.7+ with Performance Schema enabled")
                    return

                query = (
                    "SELECT enabled FROM performance_schema.setup_instruments WHERE name LIKE 'stage/innodb/alter%';"
                )
                dolphie.secondary_db_connection.execute(query)
                data = dolphie.secondary_db_connection.fetchall()
                for row in data:
                    if row.get("enabled") == "NO":
                        self.app.notify("DDL panel requires Performance Schema to have 'stage/innodb/alter%' enabled")
                        return

                self.app.toggle_panel(dolphie.panels.ddl.name)
                self.app.tab_manager.active_tab.ddl_datatable.clear()

        elif key == "7":
            if dolphie.is_mysql_version_at_least("5.7") and dolphie.performance_schema_enabled:
                if not dolphie.pfs_metrics_last_reset_time:
                    dolphie.pfs_metrics_last_reset_time = datetime.now().astimezone()
                self.app.toggle_panel(dolphie.panels.pfs_metrics.name)
            else:
                self.app.notify("Performance Schema Metrics panel requires MySQL 5.7+ with Performance Schema enabled")

        elif key == "8":
            if dolphie.is_mysql_version_at_least("5.7"):
                if dolphie.statements_summary_data:
                    dolphie.statements_summary_data.internal_data = {}
                    dolphie.statements_summary_data.filtered_data = {}
                self.app.toggle_panel(dolphie.panels.statements_summary.name)
            else:
                self.app.notify("Statements Summary panel requires MySQL 5.7+ with Performance Schema enabled")

        # Tab management commands
        elif key == "grave_accent":
            self.app.tab_manager.setup_host_tab(tab)

        elif key == "space":
            if not tab.worker or not tab.worker.is_running:
                if tab.worker_timer:
                    tab.worker_timer.stop()
                self.app.run_worker_main(tab.id)

        elif key == "plus":
            new_tab = await self.app.tab_manager.create_tab(tab_name="New Tab")
            self.app.tab_manager.switch_tab(new_tab.id)
            self.app.tab_manager.setup_host_tab(new_tab)

        elif key == "equals_sign":

            def command_get_input(tab_name):
                tab.manual_tab_name = tab_name
                self.app.tab_manager.rename_tab(tab, tab_name)

            self.app.app.push_screen(
                CommandModal(
                    command=HotkeyCommands.rename_tab,
                    message="What would you like to rename the tab to?",
                ),
                command_get_input,
            )

        elif key == "minus":
            if len(self.app.tab_manager.tabs) == 1:
                self.app.notify("Removing all tabs is not permitted", severity="error")
            else:
                if not self.app.tab_manager.active_tab:
                    self.app.notify("No active tab to remove", severity="error")
                    return

                await self.app.tab_manager.remove_tab(tab)
                await self.app.tab_manager.disconnect_tab(tab=tab, update_topbar=False)

                self.app.notify(
                    f"Tab [$highlight]{tab.name}[/$highlight] [$white]has been removed",
                    severity="success",
                )
                self.app.tab_manager.tabs.pop(tab.id, None)

        # Replay control commands
        elif key == "left_square_bracket":
            if dolphie.replay_file:
                self.app.query_one("#back_button", Button).press()

        elif key == "right_square_bracket":
            if dolphie.replay_file:
                self.app.query_one("#forward_button", Button).press()

        # Tab navigation
        elif key == "ctrl+a" or key == "ctrl+d":
            if key == "ctrl+a":
                self.app.tab_manager.host_tabs.action_previous_tab()
            elif key == "ctrl+d":
                self.app.tab_manager.host_tabs.action_next_tab()

        # Display toggle commands
        elif key == "a":
            if dolphie.show_additional_query_columns:
                dolphie.show_additional_query_columns = False
                self.app.notify("Processlist will now hide additional columns")
            else:
                dolphie.show_additional_query_columns = True
                self.app.notify("Processlist will now show additional columns")

            self.app.force_refresh_for_replay(need_current_data=True)

        elif key == "A":
            if dolphie.show_statements_summary_query_digest_text_sample:
                dolphie.show_statements_summary_query_digest_text_sample = False
                self.app.notify("Statements Summary will now show query digest")
            else:
                dolphie.show_statements_summary_query_digest_text_sample = True
                self.app.notify("Statements Summary will now show query digest sample")

        # Filter commands
        elif key == "c":
            dolphie.user_filter = None
            dolphie.db_filter = None
            dolphie.host_filter = None
            dolphie.hostgroup_filter = None
            dolphie.query_time_filter = None
            dolphie.query_filter = None

            self.app.force_refresh_for_replay(need_current_data=True)

            self.app.notify("Cleared all filters", severity="success")

        elif key == "C":
            if not dolphie.global_variables.get("innodb_thread_concurrency"):
                self.app.notify("InnoDB thread concurrency is not setup", severity="warning")
                return

            if dolphie.show_threads_with_concurrency_tickets:
                dolphie.show_threads_with_concurrency_tickets = False
                dolphie.show_idle_threads = False
                self.app.notify("Processlist will no longer only show threads with concurrency tickets")
            else:
                dolphie.show_threads_with_concurrency_tickets = True
                dolphie.show_idle_threads = True
                self.app.notify("Processlist will only show threads with concurrency tickets")

            self.app.force_refresh_for_replay(need_current_data=True)

        # Database operation commands
        elif key == "d":
            self.execute_command_in_thread(key=key)

        elif key == "D":
            await self.app.tab_manager.disconnect_tab(tab)

        elif key == "e":
            if dolphie.connection_source_alt == ConnectionSource.mariadb:
                self.app.notify(f"Command [$highlight]{key}[/$highlight] is only available for MySQL connections")
            elif dolphie.connection_source == ConnectionSource.proxysql:
                self.execute_command_in_thread(key=key)
            else:
                if dolphie.is_mysql_version_at_least("8.0") and dolphie.performance_schema_enabled:
                    self.app.app.push_screen(
                        EventLog(
                            dolphie.connection_status,
                            dolphie.app_version,
                            dolphie.host_with_port,
                            dolphie.secondary_db_connection,
                        )
                    )
                else:
                    self.app.notify("Error log command requires MySQL 8+ with Performance Schema enabled")

        elif key == "E":
            processlist = dolphie.processlist_threads_snapshot or dolphie.processlist_threads
            if processlist:
                # Extract headers from the first entry's thread_data
                first_entry = next(iter(processlist.values()))
                headers = first_entry.thread_data.keys()

                # Generate the filename with a timestamp prefix
                timestamp = datetime.now().astimezone().strftime("%Y%m%d_%H%M%S")
                filename = f"processlist-{timestamp}.csv"

                # Write the CSV to a file
                with open(filename, "w", newline="") as csvfile:
                    writer = csv.DictWriter(csvfile, fieldnames=headers)

                    # Write the headers and rows
                    writer.writeheader()
                    for process_thread in processlist.values():
                        writer.writerow(process_thread.thread_data)

                self.app.notify(
                    f"Processlist has been exported to CSV file [$highlight]{filename}",
                    severity="success",
                    timeout=10,
                )
            else:
                self.app.notify("There's no processlist data to export", severity="warning")

        elif key == "f":

            def command_get_input(filter_data):
                # Unpack the data from the modal
                filters_mapping = {
                    "User": "user_filter",
                    "Host": "host_filter",
                    "Database": "db_filter",
                    "Hostgroup": "hostgroup_filter",
                    "Minimum Query Time": "query_time_filter",
                    "Partial Query Text": "query_filter",
                }

                filters = dict(zip(filters_mapping.keys(), filter_data))

                # Apply filters and notify the user for each valid input
                for filter_name, filter_value in filters.items():
                    if filter_value:
                        if filter_name in ["Minimum Query Time", "Hostgroup"]:
                            filter_value = int(filter_value)

                        setattr(dolphie, filters_mapping[filter_name], filter_value)
                        self.app.notify(
                            f"[b]{filter_name}[/b]: [$b_highlight]{filter_value}[/$b_highlight]",
                            title="Filter applied",
                            severity="success",
                        )

                # Refresh data after applying filters
                self.app.force_refresh_for_replay(need_current_data=True)

            self.app.app.push_screen(
                CommandModal(
                    command=HotkeyCommands.thread_filter,
                    message="Filter threads by field(s)",
                    processlist_data=dolphie.processlist_threads_snapshot,
                    host_cache_data=dolphie.host_cache,
                    connection_source=dolphie.connection_source,
                ),
                command_get_input,
            )

        elif key == "i":
            if dolphie.show_idle_threads:
                dolphie.show_idle_threads = False
                dolphie.sort_by_time_descending = True

                self.app.notify("Processlist will now hide idle threads")
            else:
                dolphie.show_idle_threads = True
                dolphie.sort_by_time_descending = False

                self.app.notify("Processlist will now show idle threads")

        elif key == "k":

            def command_get_input(data):
                self.execute_command_in_thread(key=key, additional_data=data)

            self.app.app.push_screen(
                CommandModal(
                    command=HotkeyCommands.thread_kill_by_parameter,
                    message="Kill thread(s)",
                    processlist_data=dolphie.processlist_threads_snapshot,
                ),
                command_get_input,
            )

        elif key == "l" or key == "o":
            self.execute_command_in_thread(key=key)

        elif key == "M":

            def command_get_input(filter_data):
                panel = filter_data

                widget = None
                if panel == "processlist":
                    widget = tab.processlist_datatable
                elif panel == "graphs":
                    widget = tab.metric_graph_tabs
                elif panel == "metadata_locks":
                    widget = tab.metadata_locks_datatable
                elif panel == "ddl":
                    widget = tab.ddl_datatable
                elif panel == "pfs_metrics":
                    widget = tab.pfs_metrics_tabs
                elif panel == "statements_summary":
                    widget = tab.statements_summary_datatable
                elif panel == "proxysql_hostgroup_summary":
                    widget = tab.proxysql_hostgroup_summary_datatable
                elif panel == "proxysql_mysql_query_rules":
                    widget = tab.proxysql_mysql_query_rules_datatable
                elif panel == "proxysql_command_stats":
                    widget = tab.proxysql_command_stats_datatable

                if widget:
                    self.app.screen.maximize(widget)

            panel_options = [
                (panel.display_name, panel.name)
                for panel in tab.dolphie.panels.get_all_panels()
                if panel.visible and panel.name not in ["dashboard"]
            ]

            self.app.app.push_screen(
                CommandModal(
                    command=HotkeyCommands.maximize_panel,
                    maximize_panel_options=panel_options,
                    message="Maximize a Panel",
                ),
                command_get_input,
            )

        elif key == "p":
            if dolphie.replay_file:
                self.app.query_one("#pause_button", Button).press()
            else:
                if not dolphie.pause_refresh:
                    dolphie.pause_refresh = True
                    self.app.notify(f"Refresh is paused! Press [$b_highlight]{key}[/$b_highlight] again to resume")
                else:
                    dolphie.pause_refresh = False
                    self.app.notify("Refreshing has resumed", severity="success")

        if key == "P":
            if dolphie.use_performance_schema_for_processlist:
                dolphie.use_performance_schema_for_processlist = False
                self.app.notify(
                    "Switched to using [$b_highlight]Information Schema[/$b_highlight] for Processlist panel"
                )
            else:
                if dolphie.performance_schema_enabled:
                    dolphie.use_performance_schema_for_processlist = True
                    self.app.notify(
                        "Switched to using [$b_highlight]Performance Schema[/$b_highlight] for Processlist panel"
                    )
                else:
                    self.app.notify(
                        "You can't switch to [$b_highlight]Performance Schema[/$b_highlight] for "
                        "Processlist panel because it isn't enabled",
                        severity="warning",
                    )

        elif key == "q":
            self.app.app.exit()

        elif key == "r":

            def command_get_input(refresh_interval):
                dolphie.refresh_interval = refresh_interval

                self.app.notify(
                    f"Refresh interval set to [$b_highlight]{refresh_interval}[/$b_highlight] second(s)",
                    severity="success",
                )

            self.app.app.push_screen(
                CommandModal(HotkeyCommands.refresh_interval, message="Refresh Interval"),
                command_get_input,
            )

        elif key == "R":
            dolphie.metric_manager.reset()
            dolphie.reset_pfs_metrics_deltas()

            self.app.update_graphs(tab.metric_graph_tabs.get_pane(tab.metric_graph_tabs.active).name)
            dolphie.update_switches_after_reset()
            self.app.notify("Metrics have been reset", severity="success")

        elif key == "s":
            if dolphie.sort_by_time_descending:
                dolphie.sort_by_time_descending = False
                self.app.notify("Processlist will now sort threads by time in ascending order")
            else:
                dolphie.sort_by_time_descending = True
                self.app.notify("Processlist will now sort threads by time in descending order")

            self.app.force_refresh_for_replay(need_current_data=True)

        elif key == "S":
            if dolphie.replay_file:
                self.app.query_one("#seek_button", Button).press()

        elif key == "t":
            if dolphie.connection_source == ConnectionSource.proxysql:

                def command_get_input(data):
                    thread_table = Table(box=None, show_header=False)
                    thread_table.add_column("")
                    thread_table.add_column("", overflow="fold")

                    thread_id = data
                    thread_data: ProxySQLProcesslistThread = dolphie.processlist_threads_snapshot.get(thread_id)
                    if not thread_data:
                        self.app.notify(
                            f"Thread ID [$highlight]{thread_id}[/$highlight] was not found",
                            severity="error",
                        )
                        return

                    thread_table.add_row("[label]Process ID", thread_id)
                    thread_table.add_row("[label]Hostgroup", str(thread_data.hostgroup))
                    thread_table.add_row("[label]User", thread_data.user)
                    thread_table.add_row("[label]Frontend Host", thread_data.frontend_host)
                    thread_table.add_row("[label]Backend Host", thread_data.host)
                    thread_table.add_row("[label]Database", thread_data.db)
                    thread_table.add_row("[label]Command", thread_data.command)
                    thread_table.add_row("[label]Time", str(timedelta(seconds=thread_data.time)).zfill(8))

                    formatted_query = None
                    if thread_data.formatted_query.code:
                        query = sqlformat(thread_data.formatted_query.code, reindent_aligned=True)
                        formatted_query = format_query(query, minify=False)

                    self.app.app.push_screen(
                        ProxySQLThreadScreen(
                            connection_status=dolphie.connection_status,
                            app_version=dolphie.app_version,
                            host=dolphie.host_with_port,
                            thread_table=thread_table,
                            query=formatted_query,
                            extended_info=thread_data.extended_info,
                        )
                    )

            else:

                def command_get_input(data):
                    self.execute_command_in_thread(key=key, additional_data=data)

            self.app.app.push_screen(
                CommandModal(
                    command=HotkeyCommands.show_thread,
                    message="Thread Details",
                    processlist_data=dolphie.processlist_threads_snapshot,
                ),
                command_get_input,
            )

        elif key == "T":
            if dolphie.show_trxs_only:
                dolphie.show_trxs_only = False
                dolphie.show_idle_threads = False
                self.app.notify("Processlist will no longer only show threads that have an active transaction")
            else:
                dolphie.show_trxs_only = True
                dolphie.show_idle_threads = True
                self.app.notify("Processlist will only show threads that have an active transaction")

            self.app.force_refresh_for_replay(need_current_data=True)

        elif key == "u":
            if not dolphie.performance_schema_enabled and dolphie.connection_source != ConnectionSource.proxysql:
                self.app.notify("User statistics command requires Performance Schema to be enabled")
                return

            self.execute_command_in_thread(key=key)

        elif key == "V":
            global_variable_changes = tab.replay_manager.fetch_all_global_variable_changes()

            if global_variable_changes:
                table = Table(
                    box=box.SIMPLE_HEAVY,
                    show_edge=False,
                    style="table_border",
                )
                table.add_column("Timestamp")
                table.add_column("Variable")
                table.add_column("Old Value", overflow="fold")
                table.add_column("New Value", overflow="fold")

                for (
                    timestamp,
                    variable,
                    old_value,
                    new_value,
                ) in global_variable_changes:
                    table.add_row(
                        f"[dark_gray]{timestamp}",
                        f"[light_blue]{variable}",
                        old_value,
                        new_value,
                    )

                screen_data = Group(
                    Align.center(
                        "[b light_blue]Global Variable Changes[/b light_blue] "
                        f"([b highlight]{table.row_count}[/b highlight])\n"
                    ),
                    table,
                )
            else:
                self.app.notify("There are no global variable changes in this replay")

        elif key == "v":

            def command_get_input(input_variable):
                table_grid = Table.grid()
                table_counter = 1
                variable_counter = 1
                row_counter = 1
                variable_num = 1
                all_tables = []
                tables = {}
                display_global_variables = {}

                for variable, value in dolphie.global_variables.items():
                    if input_variable == "all":
                        display_global_variables[variable] = dolphie.global_variables[variable]
                    else:
                        if input_variable and input_variable in variable:
                            display_global_variables[variable] = dolphie.global_variables[variable]

                max_num_tables = 1 if len(display_global_variables) <= 50 else 2

                # Create the number of tables we want
                while table_counter <= max_num_tables:
                    tables[table_counter] = Table(box=box.HORIZONTALS, show_header=False, style="table_border")
                    tables[table_counter].add_column("")
                    tables[table_counter].add_column("")

                    table_counter += 1

                # Calculate how many global_variables per table
                row_per_count = len(display_global_variables) // max_num_tables

                # Loop global_variables
                for variable, value in display_global_variables.items():
                    tables[variable_num].add_row(f"[label]{variable}", str(value))

                    if variable_counter == row_per_count and row_counter != max_num_tables:
                        row_counter += 1
                        variable_counter = 0
                        variable_num += 1

                    variable_counter += 1

                # Put all the variable data from dict into an array
                all_tables = [table_data for table_data in tables.values() if table_data]

                # Add the data into a single tuple for add_row
                if display_global_variables:
                    table_grid.add_row(*all_tables)
                    screen_data = Align.center(table_grid)

                    self.app.app.push_screen(
                        CommandScreen(
                            dolphie.connection_status,
                            dolphie.app_version,
                            dolphie.host_with_port,
                            screen_data,
                        )
                    )
                else:
                    if input_variable:
                        self.app.notify(
                            f"No variable(s) found that match [$b_highlight]{input_variable}[/$b_highlight]"
                        )

            self.app.app.push_screen(
                CommandModal(
                    HotkeyCommands.variable_search,
                    message="Specify a variable to wildcard search",
                ),
                command_get_input,
            )

        elif key == "Z":
            if dolphie.is_mysql_version_at_least("5.7"):
                self.execute_command_in_thread(key=key)
            else:
                self.app.notify("Table size command requires MySQL 5.7+")

        elif key == "z":
            if dolphie.host_cache:
                table = Table(
                    box=box.SIMPLE_HEAVY,
                    show_edge=False,
                    style="table_border",
                )
                table.add_column("Host/IP")
                table.add_column("Hostname (if resolved)")

                for ip, addr in dolphie.host_cache.items():
                    if ip:
                        table.add_row(ip, addr)

                screen_data = Group(
                    Align.center(
                        f"[b light_blue]Host Cache[/b light_blue] "
                        f"([b highlight]{len(dolphie.host_cache)}[/b highlight])\n"
                    ),
                    table,
                )
            else:
                self.app.notify("There are currently no hosts resolved")

        if screen_data:
            self.app.app.push_screen(
                CommandScreen(
                    dolphie.connection_status,
                    dolphie.app_version,
                    dolphie.host_with_port,
                    screen_data,
                )
            )

    def execute_command_in_thread(self, key: str, additional_data=None) -> None:
        """Execute a command in a background thread.

        This method creates a daemon thread to run the command without blocking
        the UI, using call_from_thread to safely interact with Textual.

        Args:
            key: The command key that was pressed
            additional_data: Optional additional data for the command
        """

        def _run_command():
            """Internal worker function that executes in a background thread."""
            self._execute_command(key, additional_data)

        thread = threading.Thread(target=_run_command, daemon=True)
        thread.start()

    def _execute_command(self, key: str, additional_data=None) -> None:
        """Internal implementation of command execution."""
        tab = self.app.tab_manager.active_tab
        dolphie = tab.dolphie

        # These are the screens to display we use for the commands
        def show_command_screen():
            self.app.app.push_screen(
                CommandScreen(
                    dolphie.connection_status,
                    dolphie.app_version,
                    dolphie.host_with_port,
                    screen_data,
                )
            )

        def show_thread_screen():
            self.app.app.push_screen(
                ThreadScreen(
                    connection_status=dolphie.connection_status,
                    app_version=dolphie.app_version,
                    host=dolphie.host_with_port,
                    thread_table=thread_table,
                    user_thread_attributes_table=user_thread_attributes_table,
                    query=formatted_query,
                    explain_data=explain_data,
                    explain_json_data=explain_json_data,
                    explain_failure=explain_failure,
                    transaction_history_table=transaction_history_table,
                )
            )

        self.app.call_from_thread(tab.spinner.show)

        try:
            if key == "d":
                tables = {}
                all_tables = []

                db_count = dolphie.secondary_db_connection.execute(MySQLQueries.databases)
                databases = dolphie.secondary_db_connection.fetchall()

                # Determine how many tables to provide data
                max_num_tables = 1 if db_count <= 20 else 3

                # Calculate how many databases per table
                row_per_count = db_count // max_num_tables

                # Create dictionary of tables
                for table_counter in range(1, max_num_tables + 1):
                    table_box = box.HORIZONTALS
                    if max_num_tables == 1:
                        table_box = None

                    tables[table_counter] = Table(box=table_box, show_header=False, style="table_border")
                    tables[table_counter].add_column("")

                # Loop over databases
                db_counter = 1
                table_counter = 1

                # Sort the databases by name
                for database in databases:
                    tables[table_counter].add_row(database["SCHEMA_NAME"])
                    db_counter += 1

                    if db_counter > row_per_count and table_counter < max_num_tables:
                        table_counter += 1
                        db_counter = 1

                # Collect table data into an array
                all_tables = [table_data for table_data in tables.values() if table_data]

                table_grid = Table.grid()
                table_grid.add_row(*all_tables)

                screen_data = Group(
                    Align.center(f"[b light_blue]Databases[/b light_blue] ([b highlight]{db_count}[/b highlight])\n"),
                    Align.center(table_grid),
                )

                self.app.call_from_thread(show_command_screen)

            elif key == "e":
                header_style = Style(bold=True)
                table = Table(box=box.SIMPLE_HEAVY, style="table_border", show_edge=False)
                table.add_column("Hostgroup", header_style=header_style)
                table.add_column("Backend Host", max_width=35, header_style=header_style)
                table.add_column("Username", header_style=header_style)
                table.add_column("Schema", header_style=header_style)
                table.add_column("First Seen", header_style=header_style)
                table.add_column("Last Seen", header_style=header_style)
                table.add_column("Count", header_style=header_style)
                table.add_column("Error", header_style=header_style, overflow="fold")

                dolphie.secondary_db_connection.execute(ProxySQLQueries.query_errors)
                data = dolphie.secondary_db_connection.fetchall()

                for row in data:
                    table.add_row(
                        row.get("hostgroup"),
                        f"{dolphie.get_hostname(row.get('hostname'))}:{row.get('port')}",
                        row.get("username"),
                        row.get("schemaname"),
                        str(datetime.fromtimestamp(int(row.get("first_seen", 0))).astimezone()),
                        str(datetime.fromtimestamp(int(row.get("last_seen", 0))).astimezone()),
                        format_number(int(row.get("count_star", 0))),
                        "[b][highlight]{}[/b][/highlight]: {}".format(row.get("errno", 0), row.get("last_error")),
                    )

                screen_data = Group(
                    Align.center(f"[b light_blue]Query Errors ([highlight]{table.row_count}[/highlight])\n"),
                    Align.center(table),
                )

                self.app.call_from_thread(show_command_screen)

            elif key == "k":
                # Unpack the data from the modal
                (
                    kill_by_id,
                    kill_by_username,
                    kill_by_host,
                    kill_by_age_range,
                    age_range_lower_limit,
                    age_range_upper_limit,
                    kill_by_query_text,
                    include_sleeping_queries,
                ) = additional_data

                if kill_by_id:
                    try:
                        query = dolphie.build_kill_query(kill_by_id)
                        dolphie.secondary_db_connection.execute(query)

                        self.app.notify(
                            f"Killed Thread ID [$b_highlight]{kill_by_id}[/$b_highlight]",
                            severity="success",
                        )
                    except ManualException as e:
                        self.app.notify(e.reason, title="Error killing Thread ID", severity="error")
                else:
                    threads_killed = 0
                    commands_to_kill = ["Query", "Execute"]

                    if include_sleeping_queries:
                        commands_to_kill.append("Sleep")

                    # Make a copy of the threads snapshot to avoid modification during next refresh polling
                    threads = dolphie.processlist_threads_snapshot.copy()

                    for thread_id, thread in threads.items():
                        thread: ProcesslistThread
                        try:
                            # Check if the thread matches all conditions
                            if (
                                thread.command in commands_to_kill
                                and (not kill_by_username or kill_by_username == thread.user)
                                and (not kill_by_host or kill_by_host == thread.host)
                                and (
                                    not kill_by_age_range
                                    or age_range_lower_limit <= thread.time <= age_range_upper_limit
                                )
                                and (not kill_by_query_text or kill_by_query_text in thread.formatted_query.code)
                            ):
                                query = dolphie.build_kill_query(thread_id)
                                dolphie.secondary_db_connection.execute(query)

                                threads_killed += 1
                        except ManualException as e:
                            self.app.notify(
                                e.reason,
                                title=f"Error Killing Thread ID {thread_id}",
                                severity="error",
                            )

                    if threads_killed:
                        self.app.notify(f"Killed [$highlight]{threads_killed}[/$highlight] thread(s)")
                    else:
                        self.app.notify("No threads were killed")

            elif key == "l":
                status = dolphie.secondary_db_connection.fetch_value_from_field(MySQLQueries.innodb_status, "Status")
                # Extract the most recent deadlock info from the output of SHOW ENGINE INNODB STATUS
                match = re.search(
                    r"------------------------\nLATEST\sDETECTED\sDEADLOCK\n------------------------"
                    r"\n(.*?)------------\nTRANSACTIONS",
                    status,
                    flags=re.S,
                )

                if match:
                    screen_data = escape_markup(match.group(1)).replace("***", "[yellow]*****[/yellow]")
                else:
                    screen_data = Align.center("No deadlock detected")

                self.app.call_from_thread(show_command_screen)

            elif key == "o":
                screen_data = escape_markup(
                    dolphie.secondary_db_connection.fetch_value_from_field(MySQLQueries.innodb_status, "Status")
                )
                self.app.call_from_thread(show_command_screen)

            elif key == "t":
                formatted_query = ""
                explain_failure = ""
                explain_data = ""
                explain_json_data = ""

                thread_table = Table(box=None, show_header=False)
                thread_table.add_column("")
                thread_table.add_column("", overflow="fold")

                thread_id = additional_data
                thread_data: ProcesslistThread = dolphie.processlist_threads_snapshot.get(thread_id)
                if not thread_data:
                    self.app.notify(
                        f"Thread ID [$highlight]{thread_id}[/$highlight] was not found",
                        severity="error",
                    )
                    tab.spinner.hide()
                    return

                thread_table.add_row("[label]Thread ID", thread_id)
                thread_table.add_row("[label]User", thread_data.user)
                thread_table.add_row("[label]Host", thread_data.host)
                thread_table.add_row("[label]Database", thread_data.db)
                thread_table.add_row("[label]Command", thread_data.command)
                thread_table.add_row("[label]State", thread_data.state)
                thread_table.add_row("[label]Time", str(timedelta(seconds=thread_data.time)).zfill(8))
                thread_table.add_row("[label]Rows Locked", format_number(thread_data.trx_rows_locked))
                thread_table.add_row("[label]Rows Modified", format_number(thread_data.trx_rows_modified))

                thread_table.add_row("", "")
                thread_table.add_row("[label]TRX Time", thread_data.trx_time)
                thread_table.add_row("[label]TRX State", thread_data.trx_state)
                thread_table.add_row("[label]TRX Operation", thread_data.trx_operation_state)

                if thread_data.formatted_query.code:
                    query = sqlformat(thread_data.formatted_query.code, reindent_aligned=True)
                    query_db = thread_data.db

                    formatted_query = format_query(query, minify=False)

                    if query_db:
                        try:
                            dolphie.secondary_db_connection.execute(f"USE {query_db}")

                            dolphie.secondary_db_connection.execute(f"EXPLAIN {query}")
                            explain_data = dolphie.secondary_db_connection.fetchall()

                            dolphie.secondary_db_connection.execute(f"EXPLAIN FORMAT=JSON {query}")
                            explain_fetched_json_data = dolphie.secondary_db_connection.fetchone()
                            if explain_fetched_json_data:
                                explain_json_data = explain_fetched_json_data.get("EXPLAIN")
                        except ManualException as e:
                            # Error 1054 means unknown column which would result in a truncated query
                            # Error 1064 means bad syntax which would result in a truncated query
                            tip = (
                                ":bulb: [b][yellow]Tip![/b][/yellow] If the query is truncated, consider "
                                "increasing [dark_yellow]performance_schema_max_digest_length[/dark_yellow]/"
                                "[dark_yellow]max_digest_length[/dark_yellow] as a preventive measure. "
                                "If adjusting those settings isn't an option, then use command "
                                "[dark_yellow]P[/dark_yellow]. "
                                "This will switch to using SHOW PROCESSLIST instead of the Performance Schema, "
                                "which does not truncate queries.\n\n"
                                if e.code in (1054, 1064)
                                else ""
                            )

                            explain_failure = (
                                f"{tip}[b][indian_red]EXPLAIN ERROR ({e.code}):[/b] [indian_red]{e.reason}"
                            )

                user_thread_attributes_table = None
                if dolphie.performance_schema_enabled:
                    user_thread_attributes_table = Table(box=None, show_header=False, expand=True)

                    dolphie.secondary_db_connection.execute(
                        MySQLQueries.user_thread_attributes.replace("$1", thread_id)
                    )

                    user_thread_attributes = dolphie.secondary_db_connection.fetchall()
                    if user_thread_attributes:
                        user_thread_attributes_table.add_column("")
                        user_thread_attributes_table.add_column("", overflow="fold")

                        for attribute in user_thread_attributes:
                            user_thread_attributes_table.add_row(
                                f"[label]{attribute['ATTR_NAME']}",
                                attribute["ATTR_VALUE"],
                            )
                    else:
                        user_thread_attributes_table.add_column(justify="center")
                        user_thread_attributes_table.add_row("[b][label]None found")

                # Transaction history
                transaction_history_table = None
                if (
                    dolphie.is_mysql_version_at_least("5.7")
                    and dolphie.performance_schema_enabled
                    and thread_data.mysql_thread_id
                ):
                    query = MySQLQueries.thread_transaction_history.replace("$1", str(thread_data.mysql_thread_id))
                    dolphie.secondary_db_connection.execute(query)
                    transaction_history = dolphie.secondary_db_connection.fetchall()

                    if transaction_history:
                        transaction_history_table = Table(box=None)
                        transaction_history_table.add_column("Start Time")
                        transaction_history_table.add_column("Query", overflow="fold")

                        for query in transaction_history:
                            trx_history_formatted_query = ""
                            if query["sql_text"]:
                                trx_history_formatted_query = format_query(
                                    sqlformat(query["sql_text"], reindent_aligned=True),
                                    minify=False,
                                )

                            transaction_history_table.add_row(
                                query["start_time"].strftime("%Y-%m-%d %H:%M:%S"),
                                trx_history_formatted_query,
                            )

                self.app.call_from_thread(show_thread_screen)

            elif key == "u":
                if dolphie.connection_source == ConnectionSource.proxysql:
                    title = "Frontend Users"

                    dolphie.secondary_db_connection.execute(ProxySQLQueries.user_stats)
                    users = dolphie.secondary_db_connection.fetchall()

                    columns = {
                        "User": {"field": "username", "format_number": False},
                        "Active": {
                            "field": "frontend_connections",
                            "format_number": True,
                        },
                        "Max": {
                            "field": "frontend_max_connections",
                            "format_number": True,
                        },
                        "Default HG": {
                            "field": "default_hostgroup",
                            "format_number": False,
                        },
                        "Default Schema": {
                            "field": "default_schema",
                            "format_number": False,
                        },
                        "SSL": {"field": "use_ssl", "format_number": False},
                    }

                    table = Table(
                        header_style="b",
                        box=box.SIMPLE_HEAVY,
                        show_edge=False,
                        style="table_border",
                    )
                    for column, data in columns.items():
                        table.add_column(column, no_wrap=True)

                    for user in users:
                        row_values = []

                        for column, data in columns.items():
                            value = user.get(data["field"], "N/A")

                            if data["format_number"]:
                                row_values.append(format_number(value) if value else "0")
                            elif column == "SSL":
                                row_values.append("ON" if value == "1" else "OFF")
                            else:
                                row_values.append(value or "")

                        table.add_row(*row_values)
                else:
                    title = "Users"

                    if dolphie.is_mysql_version_at_least("5.7"):
                        dolphie.secondary_db_connection.execute(MySQLQueries.ps_user_statisitics)
                    else:
                        dolphie.secondary_db_connection.execute(MySQLQueries.ps_user_statisitics_56)

                    users = dolphie.secondary_db_connection.fetchall()

                    columns = {
                        "User": {"field": "user", "format_number": False},
                        "Active": {
                            "field": "current_connections",
                            "format_number": True,
                        },
                        "Total": {"field": "total_connections", "format_number": True},
                        "Rows Read": {"field": "rows_examined", "format_number": True},
                        "Rows Sent": {"field": "rows_sent", "format_number": True},
                        "Rows Updated": {
                            "field": "rows_affected",
                            "format_number": True,
                        },
                        "Tmp Tables": {
                            "field": "created_tmp_tables",
                            "format_number": True,
                        },
                        "Tmp Disk Tables": {
                            "field": "created_tmp_disk_tables",
                            "format_number": True,
                        },
                        "Plugin": {"field": "plugin", "format_number": False},
                    }

                    table = Table(
                        header_style="b",
                        box=box.SIMPLE_HEAVY,
                        show_edge=False,
                        style="table_border",
                    )
                    for column, data in columns.items():
                        table.add_column(column, no_wrap=True)

                    for user in users:
                        row_values = []

                        for column, data in columns.items():
                            value = user.get(data.get("field"), "N/A")

                            if data["format_number"]:
                                row_values.append(format_number(value) if value else "0")
                            else:
                                row_values.append(value or "")

                        table.add_row(*row_values)

                screen_data = Group(
                    Align.center(f"[b light_blue]{title} Connected ([highlight]{len(users)}[/highlight])\n"),
                    Align.center(table),
                )

                self.app.call_from_thread(show_command_screen)

            elif key == "Z":
                if dolphie.connection_source_alt == ConnectionSource.mariadb:
                    query = MySQLQueries.table_sizes_mariadb
                elif dolphie.is_mysql_version_at_least("8.0"):
                    query = MySQLQueries.table_sizes.replace("$1", "INNODB_TABLESPACES")
                else:
                    query = MySQLQueries.table_sizes.replace("$1", "INNODB_SYS_TABLESPACES")

                dolphie.secondary_db_connection.execute(query)
                database_tables_data = dolphie.secondary_db_connection.fetchall()

                columns = {
                    "Table": {"field": "DATABASE_TABLE", "format_bytes": False},
                    "Engine": {"field": "ENGINE", "format_bytes": False},
                    "Row Format": {"field": "ROW_FORMAT", "format_bytes": False},
                }

                if dolphie.connection_source_alt != ConnectionSource.mariadb:
                    columns["File Size"] = {"field": "FILE_SIZE", "format_bytes": True}
                    columns["Allocated Size"] = {"field": "ALLOCATED_SIZE", "format_bytes": True}

                columns.update(
                    {
                        "Clustered Index": {"field": "DATA_LENGTH", "format_bytes": True},
                        "Secondary Indexes": {
                            "field": "INDEX_LENGTH",
                            "format_bytes": True,
                        },
                        "Free Space": {"field": "DATA_FREE", "format_bytes": True},
                        "Frag Ratio": {
                            "field": "fragmentation_ratio",
                            "format_bytes": False,
                        },
                    }
                )

                table = Table(
                    header_style="b",
                    box=box.SIMPLE_HEAVY,
                    show_edge=False,
                    style="table_border",
                )
                for column, data in columns.items():
                    table.add_column(column, no_wrap=True)

                for database_table_row in database_tables_data:
                    row_values = []

                    for column, data in columns.items():
                        value = database_table_row.get(data.get("field"), "N/A")

                        if data["format_bytes"]:
                            row_values.append(format_bytes(value) if value else "0")
                        elif column == "Frag Ratio":
                            if value:
                                if value >= 30:
                                    row_values.append(f"[red]{value}%[/red]")
                                elif value >= 20:
                                    row_values.append(f"[yellow]{value}%[/yellow]")
                                else:
                                    row_values.append(f"[green]{value}%[/green]")
                            else:
                                row_values.append("[green]0%[/green]")

                        elif column == "Table":
                            # Color the database name. Format is database/table
                            database_table = value.split(".")
                            row_values.append(f"[dark_gray]{database_table[0]}[/dark_gray].{database_table[1]}")

                        else:
                            row_values.append(str(value) or "")

                    table.add_row(*row_values)

                screen_data = Group(
                    Align.center(
                        f"[b light_blue]Table Sizes & Fragmentation[/b light_blue] "
                        f"([highlight]{len(database_tables_data)}[/highlight])\n"
                    ),
                    Align.center(table),
                )

                self.app.call_from_thread(show_command_screen)
        except ManualException as e:
            self.app.notify(
                e.reason,
                title=f"Error running command '{key}'",
                severity="error",
                timeout=10,
            )

        self.app.call_from_thread(tab.spinner.hide)



import re

from dolphie.DataTypes import ConnectionSource, Replica
from dolphie.Modules.Functions import (
    format_bytes,
    format_number,
    format_picoseconds,
    format_time,
)
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.MySQL import Database
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Modules.TabManager import Tab
from rich.style import Style
from rich.table import Table
from textual.containers import ScrollableContainer
from textual.widgets import Static

# Example GTID: 3beacd96-6fe3-18ec-9d95-b4592zec4b45:1-26
_GTID_PATTERN = re.compile(r"\b(\w+(?:-\w+){4}):(.+)\b")


def _filter_gtid_sets(gtid_sets: str, exclude_uuids: set[str]) -> str:
    """Remove GTID lines containing any of the excluded UUIDs."""
    remaining = [
        line.strip().rstrip(",")
        for line in gtid_sets.splitlines()
        if line.strip() and not any(uuid in line for uuid in exclude_uuids)
    ]
    return ",\n".join(remaining)


def _color_gtid_sets(gtid_sets: str, primary_uuid: str) -> str:
    """Apply Rich markup to GTID sets, highlighting the primary UUID."""

    def _colorize(match):
        source_id = match.group(1)
        transaction_id = match.group(2)

        if source_id == primary_uuid:
            return f"[highlight]{source_id}[/highlight]:{transaction_id}"
        return f"[dark_gray]{source_id}:{transaction_id}[/dark_gray]"

    return _GTID_PATTERN.sub(_colorize, gtid_sets.replace(",", ""))


# MariaDB GTID format: domain_id-server_id-sequence_number (e.g., 0-1-10)
def _color_mariadb_gtid_sets(gtid_sets: str, primary_server_id) -> str:
    """Apply Rich markup to MariaDB GTID sets, highlighting the primary server_id."""
    primary_server_id = str(primary_server_id)
    colored = []
    for gtid in gtid_sets.split(","):
        gtid = gtid.strip()
        if not gtid:
            continue
        parts = gtid.split("-")
        if len(parts) >= 3 and parts[1] == primary_server_id:
            colored.append(f"[highlight]{gtid}[/highlight]")
        else:
            colored.append(f"[dark_gray]{gtid}[/dark_gray]")
    return "\n".join(colored)


def _detect_mariadb_errant_trx(replica_gtid_current_pos: str, replica_server_id, primary_gtid_current_pos: str):
    """Detect errant transactions on a MariaDB replica.

    MariaDB GTIDs encode the originating server: domain_id-server_id-sequence.
    In normal replication a replica only executes transactions from its primary, so
    its gtid_current_pos should not contain GTIDs with its own server_id
    unless someone wrote directly to it.

    We scan the replica's gtid_current_pos for entries whose server_id
    matches the replica itself, then compare against the primary's
    gtid_current_pos. A GTID is errant when:
      - The primary has no entry for that (domain_id, server_id) pair, or
      - The replica's sequence exceeds the primary's sequence.

    Returns a comma-separated string of errant GTIDs, or None if clean.
    """
    replica_server_id = str(replica_server_id)

    def _parse_gtids(gtid_str: str) -> dict:
        result = {}
        for gtid in gtid_str.split(","):
            gtid = gtid.strip()
            if not gtid:
                continue
            parts = gtid.split("-")
            if len(parts) >= 3:
                result[(parts[0], parts[1])] = int(parts[2])
        return result

    replica_gtids = _parse_gtids(replica_gtid_current_pos)
    primary_gtids = _parse_gtids(primary_gtid_current_pos)

    errant = []
    for (domain_id, server_id), seq in replica_gtids.items():
        if server_id != replica_server_id:
            continue
        primary_seq = primary_gtids.get((domain_id, server_id))
        if primary_seq is None or seq > primary_seq:
            errant.append(f"{domain_id}-{server_id}-{seq}")

    return ",".join(errant) if errant else None


def _sync_grid(grid, items: dict[str, Table], item_type: str, tab_id: str, app, tracked: dict[str, Static]):
    """Synchronize grid child widgets with the current set of items.

    Uses the caller-owned ``tracked`` dict to avoid DOM queries. The dict maps
    item keys to their mounted Static widgets and is updated in place.
    """
    current_keys = set(items.keys())

    # Update existing or mount new
    for key, table in items.items():
        if key in tracked:
            tracked[key].update(table)
        else:
            try:
                static = Static(table, id=f"{item_type}_{key}_{tab_id}")
                grid.mount(ScrollableContainer(static, id=f"{item_type}_container_{key}_{tab_id}"))
                tracked[key] = static
            except Exception:
                app.notify(
                    f"Failed to mount {item_type} [$highlight]{key}",
                    severity="error",
                )

    # Remove stale widgets
    for key in set(tracked.keys()) - current_keys:
        tracked[key].parent.remove()
        del tracked[key]


def create_panel(tab: Tab):
    dolphie = tab.dolphie

    global_variables = dolphie.global_variables
    group_replication_data = dolphie.group_replication_data
    connection_source_alt = dolphie.connection_source_alt
    panels = dolphie.panels

    # --- Replication panel ---
    if not dolphie.replication_status:
        tab.replication_container.display = False
    else:
        tab.replication_container.display = True

        if dolphie.replication_applier_status.get("data"):
            table_thread_applier_status = Table(box=None, header_style="#c5c7d2")
            table_thread_applier_status.add_column("Worker", justify="center")
            table_thread_applier_status.add_column("Usage", min_width=6)
            table_thread_applier_status.add_column("Apply Time")
            table_thread_applier_status.add_column("Last Applied Transaction")
            table_thread_applier_status.add_column("Retries")
            table_thread_applier_status.add_column("Error Time")
            table_thread_applier_status.add_column("Error Message", overflow="fold")

            for row in dolphie.replication_applier_status["data"]:
                worker_id = row.get("worker_id")
                thread_id = row.get("thread_id")

                # Handle the ROLLUP row, which contains the total for all threads
                if not thread_id:
                    all_workers_diff = dolphie.replication_applier_status["diff_all"]
                    continue

                # Calculate the difference in thread events for this worker
                worker_diff = dolphie.replication_applier_status[f"diff_{thread_id}"]

                # Format the last applied transaction
                last_applied_transaction = row.get("last_applied_transaction", "N/A")
                if last_applied_transaction and "-" in last_applied_transaction:
                    source_id_split = last_applied_transaction.split("-")[4].split(":")[0]
                    transaction_id = last_applied_transaction.split(":")[1]
                    last_applied_transaction = f"…[dark_gray]{source_id_split}[/dark_gray]:{transaction_id}"

                # Format the last error time
                last_error_time = row.get("applying_transaction_last_transient_error_timestamp", "N/A")
                last_error_time = "" if last_error_time == "0000-00-00 00:00:00.000000" else str(last_error_time)

                # Calculate the usage percentage for each worker for the current poll
                usage_percentage = round(100 * (worker_diff / all_workers_diff), 2) if all_workers_diff > 0 else 0.0
                retries_count = row.get("applying_transaction_retries_count", 0)
                retries_count = f"[dark_gray]{retries_count}" if retries_count == 0 else f"[red]{retries_count}"

                # Display the cumulative usage for the worker
                table_thread_applier_status.add_row(
                    f"[b highlight]{worker_id}[/b highlight]: {thread_id}",
                    f"{usage_percentage}%",
                    format_picoseconds(float(row["apply_time"])),
                    last_applied_transaction,
                    retries_count,
                    last_error_time,
                    row.get("applying_transaction_last_transient_error_message", "N/A"),
                )

            tab.replication_thread_applier.update(table_thread_applier_status)
            tab.replication_thread_applier_container.display = True
        else:
            tab.replication_thread_applier_container.display = False

        if connection_source_alt == ConnectionSource.mariadb:
            available_replication_variables = {
                "slave_parallel_mode": "parallel_mode",
                "slave_parallel_workers": "parallel_workers",
                "slave_parallel_threads": "parallel_threads",
                "log_slave_updates": "log_slave_updates",
            }
        elif dolphie.is_mysql_version_at_least("8.0.22"):
            available_replication_variables = {
                "replica_parallel_type": "parallel_type",
                "replica_parallel_workers": "parallel_workers",
                "replica_preserve_commit_order": "preserve_commit_order",
                "log_replica_updates": "log_replica_updates",
            }
        else:
            available_replication_variables = {
                "slave_parallel_type": "parallel_type",
                "slave_parallel_workers": "parallel_workers",
                "slave_preserve_commit_order": "preserve_commit_order",
                "log_slave_updates": "log_slave_updates",
            }

        replication_variables = "  ".join(
            f"[$label]{display_name}[/$label] {global_variables.get(var, 'N/A')}"
            for var, display_name in available_replication_variables.items()
        )

        tab.replication_variables.update(replication_variables)
        tab.replication_status.update(create_replication_table(tab))

    # --- Group Replication panel ---
    if not (dolphie.group_replication or dolphie.innodb_cluster):
        tab.group_replication_container.display = False
    else:
        tab.group_replication_container.display = True

        available_variables = {
            "group_replication_single_primary_mode": ("Single Primary", global_variables),
            "group_replication_consistency": ("Global Consistency", global_variables),
            "write_concurrency": ("Write Concurrency", group_replication_data),
        }

        group_replication_variables = "  ".join(
            f"[$label]{label}[/$label] {source.get(var, 'N/A')}" for var, (label, source) in available_variables.items()
        )

        title_prefix = panels.replication.content_key
        cluster_title = "InnoDB Cluster" if dolphie.innodb_cluster else "Group Replication"
        cluster_name = group_replication_data.get("cluster_name")
        final_cluster_name = (
            cluster_name if cluster_name else global_variables.get("group_replication_group_name", "N/A")
        )
        tab.group_replication_title.update(
            f"[b]{title_prefix}{cluster_title} ([$highlight]{final_cluster_name}[/$highlight])"
        )
        tab.group_replication_data.update(group_replication_variables)

        # Check if this server is the GR primary
        dolphie.is_group_replication_primary = any(
            row.get("MEMBER_ID") == dolphie.server_uuid and row.get("MEMBER_ROLE") == "PRIMARY"
            for row in dolphie.group_replication_members
        )

        items = create_group_replication_member_table(tab)
        _sync_grid(tab.group_replication_grid, items, "member", tab.id, dolphie.app, tab.member_widgets)

    # --- Galera Cluster panel ---
    if not dolphie.galera_cluster:
        tab.galera_container.display = False
    else:
        tab.galera_container.display = True

        gcache_size = global_variables.get("wsrep_provider_gcache_size")
        gcache_formatted = format_bytes(int(gcache_size)) if gcache_size else "N/A"

        galera_variables = (
            f"[$label]Cluster Name[/$label] {global_variables.get('wsrep_cluster_name', 'N/A')}"
            f"  [$label]SST Method[/$label] {global_variables.get('wsrep_sst_method', 'N/A')}"
            f"  [$label]OSU Method[/$label] {global_variables.get('wsrep_osu_method', 'N/A')}"
            f"  [$label]GCache Size[/$label] {gcache_formatted}"
            f"  [$label]Apply Threads[/$label] {global_variables.get('wsrep_slave_threads', 'N/A')}"
            f"  [$label]Sync Wait[/$label] {global_variables.get('wsrep_sync_wait', 'N/A')}"
        )

        title_prefix = panels.replication.content_key
        tab.galera_title.update(
            f"[b]{title_prefix}Galera Cluster"
            f" ([$highlight]{dolphie.global_status.get('wsrep_cluster_size', 'N/A')} nodes[/$highlight])"
        )
        tab.galera_data.update(galera_variables)

        items = create_galera_node_table(tab)
        _sync_grid(tab.galera_grid, items, "galera_node", tab.id, dolphie.app, tab.galera_widgets)

    # --- ClusterSet panel ---
    innodb_cluster_clustersets = dolphie.innodb_cluster_clustersets
    if not innodb_cluster_clustersets:
        tab.clusterset_container.display = False
    else:
        tab.clusterset_container.display = True

        tab.clusterset_title.update(
            f"[b]{panels.replication.content_key}ClusterSets "
            f"([$highlight]{len(innodb_cluster_clustersets)}[/$highlight])"
        )

        host_cluster_name = group_replication_data.get("cluster_name")

        items = {}
        for row in innodb_cluster_clustersets:
            clusterset_name = row["ClusterSet"]
            clusters = row["Clusters"]

            formatted_clusters = clusters.replace(host_cluster_name, f"[b highlight]{host_cluster_name}[/b highlight]")

            table = Table(box=None, show_header=False)
            table.add_column()
            table.add_column()
            table.add_row("[b][light_blue]ClusterSet", f"[light_blue]{clusterset_name}")
            table.add_row("[b][label]Clusters", formatted_clusters)

            items[clusterset_name] = table

        _sync_grid(tab.clusterset_grid, items, "clusterset", tab.id, dolphie.app, tab.clusterset_widgets)


# This function isn't in create_panel() because it's called as part of the replica worker instead of the main worker
def create_replica_panel(tab: Tab):
    dolphie = tab.dolphie

    # Refresh optimization: cache frequently accessed objects
    replica_manager = dolphie.replica_manager
    panels = dolphie.panels

    if not replica_manager.replicas:
        tab.replicas_container.display = False
        return

    tab.replicas_container.display = True
    tab.replicas_loading_indicator.display = False

    # Update replicas title
    num_replicas = len(replica_manager.available_replicas)
    title_prefix = panels.replication.content_key
    tab.replicas_title.update(f"[b]{title_prefix}Replicas ([$highlight]{num_replicas}[/$highlight])")

    # Sync replica grid widgets with current replica data
    sorted_replicas = replica_manager.get_sorted_replicas()
    items = {r.row_key: r.table for r in sorted_replicas if r.table}

    _sync_grid(tab.replicas_grid, items, "replica", tab.id, dolphie.app, tab.replica_widgets)


def create_replication_table(tab: Tab, dashboard_table=False, replica: Replica = None) -> Table:
    dolphie = tab.dolphie

    # When replica is specified, that means we're creating a table for a replica and not replication
    if replica:
        data = replica.replication_status
        mysql_version = replica.mysql_version
        connection_source_alt = replica.connection_source_alt
    else:
        data = dolphie.replication_status
        mysql_version = dolphie.host_version
        connection_source_alt = dolphie.connection_source_alt

    # Determine replication terminology based on MySQL version
    # and connection source (MariaDB or MySQL)
    source_prefix = (
        "Source"
        if dolphie.is_mysql_version_at_least("8.0.22", mysql_version)
        and connection_source_alt != ConnectionSource.mariadb
        else "Master"
    )
    replica_prefix = "Replica" if source_prefix == "Source" else "Slave"
    uuid_key = f"{source_prefix}_UUID"

    primary_uuid = data.get(uuid_key)
    primary_host = dolphie.get_hostname(data.get(f"{source_prefix}_Host"))
    primary_user = data.get(f"{source_prefix}_User")
    primary_log_file = data.get(f"{source_prefix}_Log_File")
    primary_ssl_allowed = data.get(f"{source_prefix}_SSL_Allowed")
    relay_primary_log_file = data.get(f"Relay_{source_prefix}_Log_File")
    replica_sql_running_state = data.get(f"{replica_prefix}_SQL_Running_State")
    replica_io_state = data.get(f"{replica_prefix}_IO_State")
    read_primary_log_pos = data.get(f"Read_{source_prefix}_Log_Pos")
    exec_primary_log_pos = data.get(f"Exec_{source_prefix}_Log_Pos")

    is_io_running = data.get(f"{replica_prefix}_IO_Running") == "Yes"
    is_sql_running = data.get(f"{replica_prefix}_SQL_Running") == "Yes"
    io_thread_running = "[green]ON[/green]" if is_io_running else "[red]OFF[/red]"
    sql_thread_running = "[green]ON[/green]" if is_sql_running else "[red]OFF[/red]"

    # Determine GTID status
    mariadb_using_gtid = data.get("Using_Gtid")
    mariadb_gtid_enabled = mariadb_using_gtid not in (None, "No")
    mysql_gtid_enabled = bool(data.get("Executed_Gtid_Set"))

    if mariadb_gtid_enabled:
        gtid_status = mariadb_using_gtid
    elif mysql_gtid_enabled:
        auto_position = "ON" if data.get("Auto_Position") == 1 else "OFF"
        gtid_status = f"ON [label]Auto Position[/label] {auto_position}"
    else:
        gtid_status = "OFF"

    # Replica lag calculation
    replica_lag = data.get("Seconds_Behind", 0)
    formatted_replica_lag = None
    if replica_lag is not None:
        sql_delay = data.get("SQL_Delay")
        if sql_delay:
            # Check if it's already an int or a string representing an int
            if isinstance(sql_delay, int) or (isinstance(sql_delay, str) and sql_delay.isdigit()):
                replica_lag = max(0, replica_lag - int(sql_delay))

        lag_color = "green"
        if replica_lag >= 20:
            lag_color = "red"
        elif replica_lag >= 10:
            lag_color = "yellow"

        formatted_replica_lag = f"[{lag_color}]{format_time(replica_lag)}[/{lag_color}]"

    if dashboard_table:
        table = Table(
            show_header=False,
            box=None,
            expand=True,
            title="Replication",
            title_style=Style(color="#bbc8e8", bold=True),
            style="table_border",
        )
        table.add_column(no_wrap=True)
        table.add_column(max_width=30)
    else:
        table = Table(show_header=False, box=None)
        table.add_column()
        table.add_column(overflow="fold")

    if replica:
        table.add_row("[b][light_blue]Host", f"[light_blue]{replica.host}")
        table.add_row("[b][label]Version", f"{replica.host_distro} {replica.mysql_version}")
    else:
        table.add_row("[b][label]Primary", primary_host)

    if not dashboard_table:
        table.add_row("[b][label]User", primary_user)

    table.add_row(
        "[b][label]Thread",
        f"[label]IO[/label] {io_thread_running} [label]SQL[/label] {sql_thread_running}",
    )

    replication_delay = ""
    if data["SQL_Delay"]:
        if dashboard_table:
            replication_delay = "[dark_yellow](delayed)"
        else:
            replication_delay = f"[dark_yellow]Delay[/dark_yellow] {format_time(data['SQL_Delay'])}"

    if formatted_replica_lag is None or not is_sql_running:
        table.add_row("[b][label]Lag", "")
    else:
        table.add_row(
            "[b][label]Lag",
            f"{formatted_replica_lag} [label]Speed[/label] {data['Replica_Speed']} {replication_delay}",
        )

    if dashboard_table:
        table.add_row("[label]Binlog IO", str(primary_log_file))
        table.add_row("[label]Binlog SQL", str(relay_primary_log_file))
        table.add_row("[label]Relay Log ", str(data["Relay_Log_File"]))
        table.add_row("[label]GTID", gtid_status)
        table.add_row("[label]State", str(replica_sql_running_state))
    else:
        table.add_row(
            "[b][label]Binlog IO",
            f"{primary_log_file} ([dark_gray]{read_primary_log_pos}[/dark_gray])",
        )
        table.add_row(
            "[b][label]Binlog SQL",
            f"{relay_primary_log_file} ([dark_gray]{exec_primary_log_pos}[/dark_gray])",
        )
        table.add_row(
            "[b][label]Relay Log",
            f"{data['Relay_Log_File']} ([dark_gray]{data['Relay_Log_Pos']}[/dark_gray])",
        )

    if not dashboard_table:
        ssl_enabled = "ON" if primary_ssl_allowed == "Yes" else "OFF"
        table.add_row("[b][label]SSL", ssl_enabled)

        replication_status_filtering = [
            "Replicate_Do_DB",
            "Replicate_Ignore_Table",
            "Replicate_Ignore_DB",
            "Replicate_Do_Table",
            "Replicate_Wild_Do_Table",
            "Replicate_Wild_Ignore_Table",
            "Replicate_Rewrite_DB",
        ]

        for status_filter in replication_status_filtering:
            value = data.get(status_filter)

            status_filter_formatted = f"Filter: {status_filter.split('Replicate_')[1]}"
            if value:
                table.add_row(f"[b][label]{status_filter_formatted}", str(value))

        error_types = ["Last_IO_Error", "Last_SQL_Error"]
        errors = [(error_type, data[error_type]) for error_type in error_types if data[error_type]]

        if errors:
            for error_type, error_message in errors:
                table.add_row(
                    f"[b][label]{error_type.replace('_', ' ')}",
                    f"[red]{error_message}[/red]",
                )
        else:
            table.add_row("[b][label]IO State", str(replica_io_state))
            table.add_row("[b][label]SQL State", str(replica_sql_running_state))

        if mysql_gtid_enabled:
            executed_gtid_set = data["Executed_Gtid_Set"]
            retrieved_gtid_set = data["Retrieved_Gtid_Set"]

            table.add_row("[b][label]GTID", gtid_status)

            if replica:
                # Exclude the primary's own UUID and all its replication source UUIDs to avoid
                # false positives from stale gtid_executed snapshots. The primary actively receives
                # GTIDs from its sources, so by the time replicas are checked the snapshot is behind.
                # For Group Replication, also exclude the group UUID since it has the same race.
                exclude_uuids = {dolphie.server_uuid} | dolphie.replication_source_uuids
                if dolphie.group_replication:
                    gr_group_name = dolphie.global_variables.get("group_replication_group_name")
                    if gr_group_name:
                        exclude_uuids.add(gr_group_name)

                replica_gtid_set = _filter_gtid_sets(executed_gtid_set, exclude_uuids)
                primary_gtid_set = _filter_gtid_sets(dolphie.global_variables.get("gtid_executed", ""), exclude_uuids)

                replica.connection.execute(
                    "SELECT GTID_SUBTRACT(%s, %s) AS errant_trxs",
                    (replica_gtid_set, primary_gtid_set),
                )
                gtid_data = replica.connection.fetchone()
                if gtid_data.get("errant_trxs"):
                    errant_trx = f"[red]{gtid_data['errant_trxs']}[/red]"
                else:
                    errant_trx = "[green]None[/green]"

                table.add_row("[b][label]Errant TRX", errant_trx)

                # If this replica has replicas, use its primary server UUID, else use its own
                primary_uuid = primary_uuid or dolphie.server_uuid

            retrieved_gtid_set = _color_gtid_sets(retrieved_gtid_set, primary_uuid)
            executed_gtid_set = _color_gtid_sets(executed_gtid_set, primary_uuid)

            table.add_row("[b][label]Retrieved GTID", retrieved_gtid_set)
            table.add_row("[b][label]Executed GTID", executed_gtid_set)
        elif mariadb_gtid_enabled:
            primary_id = data.get("Master_Server_Id")

            table.add_row("[b][label]GTID", gtid_status)

            if replica:
                # Determine the primary server ID for coloring
                replica_primary_server_id = dolphie.replication_status.get("Master_Server_Id")
                primary_id = replica_primary_server_id or dolphie.global_variables.get("server_id")

                replica.connection.execute(
                    "SELECT @@server_id AS server_id, @@gtid_slave_pos AS gtid_slave_pos, "
                    "@@gtid_current_pos AS gtid_current_pos"
                )
                gtid_data = replica.connection.fetchone()

                replica_server_id = gtid_data.get("server_id")
                replica_gtid_slave_pos = gtid_data.get("gtid_slave_pos", "")
                replica_gtid_current_pos = gtid_data.get("gtid_current_pos", "")

                # Detect errant transactions
                primary_gtid_current_pos = dolphie.global_variables.get("gtid_current_pos", "")
                if replica_gtid_current_pos and primary_gtid_current_pos:
                    errant = _detect_mariadb_errant_trx(
                        replica_gtid_current_pos, replica_server_id, primary_gtid_current_pos
                    )
                    errant_trx = f"[red]{errant}[/red]" if errant else "[green]None[/green]"
                else:
                    errant_trx = "[green]None[/green]"
                table.add_row("[b][label]Errant TRX", errant_trx)

                # Retrieved GTID from SHOW SLAVE STATUS
                gtid_io_pos = data.get("Gtid_IO_Pos")
                if gtid_io_pos:
                    table.add_row("[b][label]Retrieved GTID", _color_mariadb_gtid_sets(gtid_io_pos, primary_id))

                # Executed GTID from the replica's gtid_slave_pos
                if replica_gtid_slave_pos:
                    table.add_row(
                        "[b][label]Executed GTID", _color_mariadb_gtid_sets(replica_gtid_slave_pos, primary_id)
                    )
            else:
                # Self-view: this host is a replica
                gtid_io_pos = data.get("Gtid_IO_Pos")
                if gtid_io_pos:
                    table.add_row("[b][label]Retrieved GTID", _color_mariadb_gtid_sets(gtid_io_pos, primary_id))

                gtid_slave_pos = dolphie.global_variables.get("gtid_slave_pos")
                if gtid_slave_pos:
                    table.add_row("[b][label]Executed GTID", _color_mariadb_gtid_sets(gtid_slave_pos, primary_id))

    return table


def create_group_replication_member_table(tab: Tab) -> dict[str, Table]:
    dolphie = tab.dolphie

    if not dolphie.group_replication_members:
        return {}

    unsorted: list[tuple[str, str, Table]] = []  # (member_id, host, table)
    for row in dolphie.group_replication_members:
        member_id = row.get("MEMBER_ID")
        member_host = row.get("MEMBER_HOST")
        member_port = row.get("MEMBER_PORT")
        host = f"{member_host}:{member_port}"

        member_role = row.get("MEMBER_ROLE", "N/A")
        if member_role == "PRIMARY":
            member_role = f"[b][highlight]{member_role}[/highlight]"

        member_state = row.get("MEMBER_STATE", "N/A")
        member_state = f"[green]{member_state}[/green]" if member_state == "ONLINE" else f"[red]{member_state}[/red]"

        table = Table(box=None, show_header=False)
        table.add_column()
        table.add_column()

        table.add_row("[b][light_blue]Member", f"[light_blue]{host}")
        table.add_row("[b][label]UUID", str(member_id))
        table.add_row("[b][label]Role", member_role)
        table.add_row("[b][label]State", member_state)
        table.add_row("[b][label]Version", row.get("MEMBER_VERSION", "N/A"))

        table.add_row(
            "[b][label]Certifier",
            f"[label]Queue[/label] {format_number(row.get('COUNT_TRANSACTIONS_IN_QUEUE', 'N/A'))}"
            f" [label]Checked[/label] {format_number(row.get('COUNT_TRANSACTIONS_CHECKED', 'N/A'))}"
            f" [label]Detected[/label] {format_number(row.get('COUNT_CONFLICTS_DETECTED', 'N/A'))}",
        )
        table.add_row(
            "[b][label]Applier",
            f"{format_number(row.get('COUNT_TRANSACTIONS_REMOTE_APPLIED', 'N/A'))}"
            f" [label]Queue[/label] {format_number(row.get('COUNT_TRANSACTIONS_REMOTE_IN_APPLIER_QUEUE', 'N/A'))}",
        )
        table.add_row(
            "[b][label]Local",
            f"[label]Proposed[/label] {format_number(row.get('COUNT_TRANSACTIONS_LOCAL_PROPOSED', 'N/A'))}"
            f" [label]Rollback[/label] {format_number(row.get('COUNT_TRANSACTIONS_LOCAL_ROLLBACK', 'N/A'))}",
        )
        table.add_row(
            "[b][label]Cert Rows",
            format_number(row.get("COUNT_TRANSACTIONS_ROWS_VALIDATING", "N/A")),
        )

        unsorted.append((member_id, host, table))

    # Return sorted by host, keyed by member_id
    return {mid: tbl for mid, _, tbl in sorted(unsorted, key=lambda x: x[1])}


def create_galera_node_table(tab: Tab) -> dict[str, Table]:
    dolphie = tab.dolphie
    galera = dolphie.global_status

    if not dolphie.galera_cluster_members:
        return {}

    local_uuid = galera.get("wsrep_gcomm_uuid")

    unsorted: list[tuple[str, str, Table]] = []
    for row in dolphie.galera_cluster_members:
        node_uuid = row.get("node_uuid")
        node_name = row.get("node_name", "N/A")
        node_address = row.get("node_incoming_address", "N/A")
        # Strip the port if it's 0 (default when wsrep-node-incoming-address isn't set)
        if node_address.endswith(":0"):
            node_address = node_address[:-2]
        is_local = node_uuid == local_uuid

        table = Table(box=None, show_header=False)
        table.add_column()
        table.add_column()

        if is_local:
            table.add_row(
                "[b][light_blue]Member", f"[b][highlight]{node_name}[/highlight] [light_blue]({node_address})"
            )
        else:
            table.add_row("[b][light_blue]Member", f"[light_blue]{node_name} ({node_address})")

        table.add_row("[b][label]UUID", f"[dark_gray]{node_uuid}[/dark_gray]")

        if is_local:
            # Node state
            node_state = galera.get("wsrep_local_state_comment", "N/A")
            if node_state == "Synced":
                node_state_colored = f"[green]{node_state}[/green]"
            elif node_state in ("Donor/Desynced", "Donor"):
                node_state_colored = f"[yellow]{node_state}[/yellow]"
            else:
                node_state_colored = f"[red]{node_state}[/red]"

            table.add_row(
                "[b][label]State",
                f"{node_state_colored}"
                f"  [label]Connected[/label] {galera.get('wsrep_connected', 'N/A')}"
                f"  [label]Ready[/label] {galera.get('wsrep_ready', 'N/A')}",
            )

            # Flow control
            flow_control_paused = float(galera.get("wsrep_flow_control_paused", 0))
            recv_q = float(galera.get("wsrep_local_recv_queue_avg", 0))
            send_q = float(galera.get("wsrep_local_send_queue_avg", 0))

            table.add_row(
                "[b][label]Flow Control",
                f"[label]Paused[/label] {flow_control_paused:.4f}"
                f"  [label]Recv Q[/label] {recv_q:.4f}"
                f"  [label]Send Q[/label] {send_q:.4f}",
            )

            # Certification
            cert_failures = int(galera.get("wsrep_local_cert_failures", 0))
            bf_aborts = int(galera.get("wsrep_local_bf_aborts", 0))

            table.add_row(
                "[b][label]Certification",
                f"[label]Deps[/label] {format_number(galera.get('wsrep_cert_deps_distance', 0))}"
                f"  [label]Failures[/label] {format_number(cert_failures)}"
                f"  [label]Aborts[/label] {format_number(bf_aborts)}",
            )
            table.add_row(
                "[b][label]Writesets",
                f"[label]Replicated[/label] {format_number(galera.get('wsrep_replicated', 0))}"
                f" ({format_bytes(galera.get('wsrep_replicated_bytes', 0))})"
                f"  [label]Received[/label] {format_number(galera.get('wsrep_received', 0))}"
                f" ({format_bytes(galera.get('wsrep_received_bytes', 0))})",
            )

        unsorted.append((node_uuid, node_name, table))

    # Return sorted by node_name, keyed by node_uuid
    return {uid: tbl for uid, _, tbl in sorted(unsorted, key=lambda x: x[1])}


def fetch_replication_data(tab: Tab, replica: Replica = None) -> dict:
    dolphie = tab.dolphie
    connection = replica.connection if replica else dolphie.main_db_connection
    mysql_version = replica.mysql_version if replica else None
    connection_source_alt = replica.connection_source_alt if replica else dolphie.connection_source_alt

    # Determine replication status query
    use_show_replica_status = (
        dolphie.is_mysql_version_at_least("8.0.22", use_version=mysql_version)
        and connection_source_alt != ConnectionSource.mariadb
    )
    replication_status_query = (
        MySQLQueries.show_replica_status if use_show_replica_status else MySQLQueries.show_slave_status
    )

    # Determine lag source and query
    replica_lag_source = "HB" if dolphie.heartbeat_table else None
    replica_lag_query = MySQLQueries.heartbeat_replica_lag if replica_lag_source else replication_status_query

    # Fetch replication status
    connection.execute(replication_status_query)
    all_rows = connection.fetchall()
    replication_status = all_rows[0] if all_rows else {}

    # Collect all source UUIDs for multi-source replication errant TRX detection
    if not replica and all_rows:
        uuid_field = "Source_UUID" if use_show_replica_status else "Master_UUID"
        dolphie.replication_source_uuids = {row.get(uuid_field) for row in all_rows if row.get(uuid_field)}

    # Fetch replica lag using alternative method if applicable
    if replica_lag_source:
        connection.execute(replica_lag_query)
        replica_lag_data = connection.fetchone()
    else:
        replica_lag_data = replication_status

    # Determine lag key and calculate replica lag
    lag_key = "Seconds_Behind_Source" if use_show_replica_status else "Seconds_Behind_Master"
    seconds_behind = replica_lag_data.get(lag_key)
    replica_lag = int(seconds_behind) if seconds_behind is not None else 0

    if replication_status:
        # Update replication status with lag and speed
        previous_lag = (
            replica.replication_status.get("Seconds_Behind", 0)
            if replica
            else dolphie.replication_status.get("Seconds_Behind", 0)
        )
        replication_status["Seconds_Behind"] = replica_lag
        replication_status["Replica_Speed"] = (
            round((previous_lag - replica_lag) / dolphie.polling_latency)
            if previous_lag and replica_lag < previous_lag and dolphie.polling_latency > 0
            else 0
        )

    return replication_status


def fetch_replicas(tab: Tab):
    dolphie = tab.dolphie

    # Track which row_keys are active this cycle so we can remove stale ones after
    active_row_keys = set()

    # Build lookup for existing replicas by thread_id for MariaDB port reuse
    replica_by_thread_id = (
        {r.thread_id: r for r in dolphie.replica_manager.replicas.values()}
        if dolphie.connection_source_alt == ConnectionSource.mariadb
        else {}
    )

    for row in dolphie.replica_manager.available_replicas:
        replica_error = None
        host = dolphie.get_hostname(row["host"].split(":")[0])

        if dolphie.connection_source_alt == ConnectionSource.mariadb:
            # MariaDB: no UUID for correlation - check if we already have this replica
            # by thread_id, otherwise rotate through available ports
            existing = replica_by_thread_id.get(row.get("id"))
            if existing:
                assigned_port = existing.port
                host = existing.host.split(":")[0]
            else:
                assigned_port = None
                for port_data in dolphie.replica_manager.ports.values():
                    if not port_data.get("in_use"):
                        assigned_port = port_data["port"]
                        port_data["in_use"] = True

                        # Use report_host from SHOW SLAVE HOSTS if specified, otherwise
                        # fall back to the processlist IP
                        report_host = port_data.get("host")
                        if report_host:
                            host = dolphie.get_hostname(report_host)

                        break
        else:
            # MySQL: correlate the replica in the processlist to a specific port from
            # SHOW REPLICAS via UUID — processlist IP is used for the host
            assigned_port = dolphie.replica_manager.ports.get(row.get("replica_uuid"), {}).get("port", 3306)

        # Update the port of available_replicas so it can be used for the row_key
        row["port"] = assigned_port
        port = assigned_port

        row_key = dolphie.replica_manager.create_replica_row_key(row.get("host"), port)
        active_row_keys.add(row_key)

        host_and_port = f"{host}:{port}" if port else host

        replica = dolphie.replica_manager.get_replica(row_key)
        if not replica:
            replica = dolphie.replica_manager.add_replica(
                row_key=row_key, thread_id=row.get("id"), host=host_and_port, port=port
            )

        # If we don't have a replica connection, we create one
        if not replica.connection and port:
            try:
                replica.connection = Database(
                    app=dolphie.app,
                    host=host,
                    user=dolphie.user,
                    password=dolphie.password,
                    port=port,
                    socket=None,
                    ssl=dolphie.ssl,
                    save_connection_id=False,
                )
                global_variables = replica.connection.fetch_status_and_variables("variables")

                replica.mysql_version = dolphie.parse_server_version(global_variables.get("version"))
                replica.host_distro, replica.connection_source_alt = dolphie.determine_distro_and_connection_source_alt(
                    global_variables
                )
            except ManualException as e:
                replica_error = e.reason

        # If we have a replica connection, we fetch its replication status
        if replica.connection:
            try:
                replica.replication_status = fetch_replication_data(tab, replica)
                if replica.replication_status:
                    replica.table = create_replication_table(tab, replica=replica)
            except ManualException as e:
                replica_error = e.reason

        if replica_error:
            table = Table(box=None, show_header=False)
            table.add_column()
            table.add_column(overflow="fold")

            table.add_row("[b][light_blue]Host", f"[light_blue]{host_and_port}")
            table.add_row("[b][label]User", row["user"])
            table.add_row("[b][label]Error", f"[red]{replica_error}")

            replica.table = table

    # Remove replicas that are no longer in available_replicas
    for row_key in set(dolphie.replica_manager.replicas.keys()) - active_row_keys:
        dolphie.replica_manager.remove_replica(row_key)



from dolphie.DataTypes import ProcesslistThread
from dolphie.Modules.Functions import format_number, format_query
from dolphie.Modules.Queries import MySQLQueries
from dolphie.Modules.TabManager import Tab
from rich.syntax import Syntax
from textual.widgets import DataTable


def create_panel(tab: Tab) -> DataTable:
    dolphie = tab.dolphie

    columns = [
        {"name": "Thread ID", "field": "id", "width": None, "format_number": False},
    ]

    if dolphie.use_performance_schema_for_processlist:
        columns.extend(
            [
                {
                    "name": "Protocol",
                    "field": "protocol",
                    "width": 8,
                    "format_number": False,
                }
            ]
        )

    columns.extend(
        [
            {"name": "Username", "field": "user", "width": 20, "format_number": False},
        ]
    )

    if dolphie.show_additional_query_columns:
        columns.extend(
            [
                {
                    "name": "Hostname/IP",
                    "field": "host",
                    "width": 25,
                    "format_number": False,
                },
                {
                    "name": "Database",
                    "field": "db",
                    "width": 15,
                    "format_number": False,
                },
            ]
        )

    columns.extend(
        [
            {"name": "Command", "field": "command", "width": 8, "format_number": False},
            {"name": "State", "field": "state", "width": 20, "format_number": False},
            {
                "name": "TRX State",
                "field": "trx_state",
                "width": 9,
                "format_number": False,
            },
            {
                "name": "R-Lock",
                "field": "trx_rows_locked",
                "width": 7,
                "format_number": True,
            },
            {
                "name": "R-Mod",
                "field": "trx_rows_modified",
                "width": 7,
                "format_number": True,
            },
        ]
    )

    if (
        dolphie.show_additional_query_columns and dolphie.global_variables.get("innodb_thread_concurrency")
    ) or dolphie.show_threads_with_concurrency_tickets:
        columns.append(
            {
                "name": "Tickets",
                "field": "trx_concurrency_tickets",
                "width": 8,
                "format_number": False,
            }
        )

    if dolphie.show_trxs_only:
        columns.append(
            {
                "name": "TRX Age",
                "field": "trx_time",
                "width": 9,
                "format_number": False,
            },
        )

    columns.extend(
        [
            {
                "name": "Age",
                "field": "formatted_time",
                "width": 9,
                "format_number": False,
            },
            {
                "name": "Query",
                "field": "formatted_query",
                "width": None,
                "format_number": False,
            },
            {
                "name": "time_seconds",
                "field": "time",
                "width": 0,
                "format_number": False,
            },
        ]
    )

    query_length_max = 300
    processlist_datatable = tab.processlist_datatable

    if len(processlist_datatable.columns) != len(columns):
        processlist_datatable.clear(columns=True)

    column_names = []
    column_fields = []
    column_format_numbers = []

    if not processlist_datatable.columns:
        for column_data in columns:
            processlist_datatable.add_column(column_data["name"], key=column_data["name"], width=column_data["width"])

    for column_data in columns:
        column_names.append(column_data["name"])
        column_fields.append(column_data["field"])
        column_format_numbers.append(column_data["format_number"])

    threads_to_render: dict[str, ProcesslistThread] = {}
    # We use filter here for replays since the original way requires changing WHERE clause
    if dolphie.replay_file:
        for thread_id, thread in dolphie.processlist_threads.items():
            thread: ProcesslistThread

            # Check each filter condition and skip thread if it doesn't match
            if dolphie.show_trxs_only and thread.trx_state == "[dark_gray]N/A":
                continue

            if dolphie.user_filter and dolphie.user_filter != thread.user:
                continue

            if dolphie.db_filter and dolphie.db_filter != thread.db:
                continue

            if dolphie.host_filter and dolphie.host_filter not in thread.host:
                continue

            if dolphie.query_time_filter and thread.time < dolphie.query_time_filter:
                continue

            if dolphie.query_filter and dolphie.query_filter not in thread.formatted_query.code:
                continue

            if dolphie.show_threads_with_concurrency_tickets and thread.trx_concurrency_tickets == "[dark_gray]0":
                continue

            # If all checks passed, add it to the visible list
            threads_to_render[thread_id] = thread
    else:
        # Not a replay file, so fetch_data() already filtered.
        threads_to_render = dolphie.processlist_threads

    changed = False

    with dolphie.app.batch_update():
        # Remove stale rows first so updates/adds operate on a cleaner table
        if threads_to_render:
            rows_to_remove = set(processlist_datatable.rows.keys()) - set(threads_to_render.keys())
            if rows_to_remove:
                changed = True
                # Bulk-clear and re-add when most rows are stale
                if len(rows_to_remove) > len(threads_to_render):
                    processlist_datatable.clear()
                else:
                    for id in rows_to_remove:
                        processlist_datatable.remove_row(id)
        else:
            # No threads to show, clear the table
            if processlist_datatable.row_count:
                changed = True
                processlist_datatable.clear()

        for thread_id, thread in threads_to_render.items():
            thread: ProcesslistThread

            row_values = []
            for column_field, column_format_number in zip(column_fields, column_format_numbers):
                value = getattr(thread, column_field)
                if column_format_number:
                    value = format_number(value)
                if column_field == "formatted_query" and isinstance(value, Syntax):
                    value = format_query(value.code[:query_length_max])
                row_values.append(value)

            if thread_id in processlist_datatable.rows:
                datatable_row = processlist_datatable.get_row(thread_id)

                for column_id, (column_name, column_field) in enumerate(zip(column_names, column_fields)):
                    new_val = row_values[column_id]
                    old_val = datatable_row[column_id]

                    # Compare text content for Syntax objects
                    cmp_new = new_val.code if isinstance(new_val, Syntax) else new_val
                    cmp_old = old_val.code if isinstance(old_val, Syntax) else old_val

                    if cmp_new != cmp_old or column_field == "formatted_time" or column_field == "time":
                        changed = True
                        processlist_datatable.update_cell(
                            thread_id,
                            column_name,
                            new_val,
                            update_width=(column_field == "formatted_query"),
                        )
            else:
                changed = True
                processlist_datatable.add_row(*row_values, key=thread_id)

        if changed:
            processlist_datatable.sort("time_seconds", reverse=dolphie.sort_by_time_descending)

    if dolphie.replay_file:
        dolphie.processlist_threads = threads_to_render

    title = f"{dolphie.panels.processlist.title} " f"([$highlight]{processlist_datatable.row_count}[/$highlight]"
    if dolphie.show_threads_with_concurrency_tickets:
        title += f"/[$highlight]{dolphie.global_variables.get('innodb_thread_concurrency')}[/$highlight]"
    title += ")"
    tab.processlist_title.update(title)


def fetch_data(tab: Tab) -> dict[str, ProcesslistThread]:
    dolphie = tab.dolphie

    # Determine query and column names based on whether performance_schema is used
    if dolphie.performance_schema_enabled and dolphie.use_performance_schema_for_processlist:
        processlist_query = MySQLQueries.ps_query
        if not dolphie.is_mysql_version_at_least("5.7"):
            processlist_query = processlist_query.replace("connection_type", '""')
        user_col, db_col, host_col, time_col, info_col, state_col, command_col = (
            "processlist_user",
            "processlist_db",
            "processlist_host",
            "processlist_time",
            "processlist_info",
            "processlist_state",
            "processlist_command",
        )
    else:
        processlist_query = MySQLQueries.pl_query
        user_col, db_col, host_col, time_col, info_col, state_col, command_col = (
            "User",
            "db",
            "Host",
            "Time",
            "Info",
            "State",
            "Command",
        )

    # Build the WHERE clause
    where_clause = []
    if not dolphie.show_idle_threads:
        where_clause.append(
            f"({command_col} != 'Sleep' AND {command_col} NOT LIKE 'Binlog Dump%') AND ({info_col}"
            f" IS NOT NULL OR trx_query IS NOT NULL) AND IFNULL({state_col}, '') NOT LIKE 'Group Replication"
            " Module%'"
        )
    if dolphie.show_trxs_only:
        where_clause.append("trx_state != ''")
    if dolphie.show_threads_with_concurrency_tickets:
        where_clause.append("trx_concurrency_tickets > 0")
    if dolphie.user_filter:
        where_clause.append(f"{user_col} = '{dolphie.user_filter}'")
    if dolphie.db_filter:
        where_clause.append(f"{db_col} = '{dolphie.db_filter}'")
    if dolphie.host_filter:
        where_clause.append(f"{host_col} LIKE '{dolphie.host_filter}%'")
    if dolphie.query_time_filter:
        where_clause.append(f"{time_col} >= '{dolphie.query_time_filter}'")
    if dolphie.query_filter:
        if dolphie.use_performance_schema_for_processlist:
            where_clause.append(
                f"({info_col} LIKE '%%{dolphie.query_filter}%%' OR trx_query LIKE '%%{dolphie.query_filter}%%')"
            )
        else:
            where_clause.append(f"{info_col} LIKE '%%{dolphie.query_filter}%%'")

    # Add the WHERE clause to the query
    if where_clause:
        processlist_query = processlist_query.replace("$1", "AND " + " AND ".join(where_clause))
    else:
        processlist_query = processlist_query.replace("$1", "")

    # Execute the query and fetch the results
    dolphie.main_db_connection.execute(processlist_query)
    threads = dolphie.main_db_connection.fetchall()

    processlist_threads = {}
    for thread in threads:
        # Don't include Dolphie's own threads
        if dolphie.main_db_connection.connection_id == thread["id"] or (
            dolphie.secondary_db_connection and dolphie.secondary_db_connection.connection_id == thread["id"]
        ):
            continue

        # Use trx_query from InnoDB since it's more accurate than P_S
        if dolphie.use_performance_schema_for_processlist and thread["trx_query"]:
            thread["query"] = thread["trx_query"]
        thread["query"] = thread["query"] or ""

        # Resolve hostname if possible
        if thread["host"]:
            host = thread["host"].split(":")[0]
            thread["host"] = dolphie.get_hostname(host)

        # We don't need trx_query anymore
        thread.pop("trx_query", None)

        processlist_threads[str(thread["id"])] = ProcesslistThread(thread)

    return processlist_threads



from collections import deque
from datetime import datetime
from functools import partial
from typing import TYPE_CHECKING

from textual.worker import Worker, WorkerState, get_current_worker

import dolphie.Modules.MetricManager as MetricManager
from dolphie.DataTypes import ConnectionSource, ConnectionStatus
from dolphie.Modules.ManualException import ManualException
from dolphie.Modules.ReplayManager import ReplayManager
from dolphie.Panels import Replication as ReplicationPanel

if TYPE_CHECKING:
    from dolphie.App import DolphieApp


class WorkerManager:
    """This module handles all worker management operations.

    This includes main refresh worker, replicas worker, and replay worker, along with
    their state change handlers.
    """

    def __init__(self, app: "DolphieApp"):
        """Initialize the WorkerManager.

        Args:
            app: Reference to the main DolphieApp instance
        """
        self.app = app

    async def run_worker_replay(self, tab_id: str, manual_control: bool = False):
        tab = self.app.tab_manager.get_tab(tab_id)
        if not tab:
            return

        try:
            # Get our worker thread
            tab.worker = get_current_worker()
            tab.worker.name = tab_id

            dolphie = tab.dolphie

            tab.replay_manual_control = manual_control
            if (
                len(self.app.screen_stack) > 1
                or (dolphie.pause_refresh and not manual_control)
                or tab.id != self.app.tab_manager.active_tab.id
            ):
                return

            # Get the next event from the replay file
            replay_event_data = tab.replay_manager.get_next_refresh_interval()
            # If there's no more events, stop here and cancel the worker
            if not replay_event_data:
                tab.worker.cancel()

                return

            tab.replay_manager.fetch_global_variable_changes_for_current_replay_id()

            # Common data for refreshing
            dolphie.system_utilization = replay_event_data.system_utilization
            dolphie.global_variables = replay_event_data.global_variables
            dolphie.global_status = replay_event_data.global_status
            common_metrics = {
                "system_utilization": dolphie.system_utilization,
                "global_variables": dolphie.global_variables,
                "global_status": dolphie.global_status,
            }

            dolphie.worker_processing_time = dolphie.global_status.get("replay_polling_latency", 0)

            if dolphie.connection_source == ConnectionSource.mysql:
                dolphie.host_version = dolphie.parse_server_version(dolphie.global_variables.get("version"))
                dolphie.binlog_status = replay_event_data.binlog_status
                dolphie.innodb_metrics = replay_event_data.innodb_metrics
                dolphie.replica_manager.available_replicas = replay_event_data.replica_manager
                dolphie.processlist_threads = replay_event_data.processlist
                dolphie.replication_status = replay_event_data.replication_status
                dolphie.replication_applier_status = replay_event_data.replication_applier_status
                dolphie.metadata_locks = replay_event_data.metadata_locks
                dolphie.group_replication_members = replay_event_data.group_replication_members
                dolphie.group_replication_data = replay_event_data.group_replication_data
                dolphie.galera_cluster_members = replay_event_data.galera_cluster_members
                dolphie.file_io_data = replay_event_data.file_io_data
                dolphie.table_io_waits_data = replay_event_data.table_io_waits_data
                dolphie.statements_summary_data = replay_event_data.statements_summary_data

                dolphie.pfs_metrics_last_reset_time = dolphie.global_status.get("pfs_metrics_last_reset_time", 0)

                connection_source_metrics = {
                    "innodb_metrics": dolphie.innodb_metrics,
                    "replication_status": dolphie.replication_status,
                }

                if not dolphie.server_uuid:
                    dolphie.configure_mysql_variables()
            elif dolphie.connection_source == ConnectionSource.proxysql:
                dolphie.host_version = dolphie.parse_server_version(dolphie.global_variables.get("admin-version"))
                dolphie.proxysql_command_stats = replay_event_data.command_stats
                dolphie.proxysql_hostgroup_summary = replay_event_data.hostgroup_summary
                dolphie.processlist_threads = replay_event_data.processlist

                connection_source_metrics = {"proxysql_command_stats": dolphie.proxysql_command_stats}

            # Refresh the metric manager metrics to the state of the replay event
            dolphie.metric_manager.refresh_data(
                worker_start_time=datetime.now().astimezone(),
                **common_metrics,
                **connection_source_metrics,
            )

            # Metrics data is already calculated in the replay event data so we just need to update the values
            is_delta_metrics = replay_event_data.metric_manager.get("_delta", False)
            new_datetimes = replay_event_data.metric_manager.get("datetimes", [])

            if is_delta_metrics:
                # Delta format from daemon mode
                new_dt = new_datetimes[0] if new_datetimes else None
                last_dt = dolphie.metric_manager.datetimes[-1] if dolphie.metric_manager.datetimes else None

                is_sequential = False
                if last_dt is not None and new_dt is not None and new_dt > last_dt:
                    # Verify the gap isn't too large (e.g. daemon restart after downtime)
                    try:
                        dt_fmt = MetricManager.MetricManager.DATETIME_FORMAT
                        new_dt_parsed = datetime.strptime(new_dt, dt_fmt)
                        gap = (new_dt_parsed - datetime.strptime(last_dt, dt_fmt)).total_seconds()
                        is_sequential = gap <= dolphie.refresh_interval * 10
                    except ValueError:
                        new_dt_parsed = None

                if is_sequential:
                    # Normal forward step: append new delta value
                    dolphie.metric_manager.datetimes.append(new_dt)
                    for metric_name, metric_data in replay_event_data.metric_manager.items():
                        metric_instance = dolphie.metric_manager.metrics.__dict__.get(metric_name)
                        if metric_instance:
                            for field_name, metric_values in metric_data.items():
                                metric: MetricManager.MetricData = metric_instance.__dict__.get(field_name)
                                if metric and metric_values:
                                    metric.values.append(metric_values[0])
                                    metric.last_value = metric_values[0]

                    # Trim to rolling window
                    if new_dt_parsed:
                        dolphie.metric_manager.trim_datetimes_to_window(new_dt_parsed)
                else:
                    # Non-sequential (backward, seek, or first event): rebuild the rolling window
                    metrics_list = tab.replay_manager.fetch_delta_metrics_for_window(
                        tab.replay_manager.current_replay_id
                    )

                    # Old format entries contain complete snapshots so skip to the last one
                    # and only accumulate delta entries after it
                    last_full_idx = -1
                    for i, entry in enumerate(metrics_list):
                        if not entry.get("_delta", False):
                            last_full_idx = i
                    if last_full_idx >= 0:
                        metrics_list = metrics_list[last_full_idx:]

                    # Clear existing metrics
                    dolphie.metric_manager.datetimes.clear()
                    for metric_data_obj in dolphie.metric_manager._all_metrics_data_history:
                        metric_data_obj.values.clear()

                    # Rebuild from the window entries
                    for entry in metrics_list:
                        entry_datetimes = entry.get("datetimes", [])
                        if entry_datetimes:
                            dolphie.metric_manager.datetimes.extend(entry_datetimes)

                        for metric_name, metric_data in entry.items():
                            if metric_name in ("datetimes", "_delta"):
                                continue
                            metric_instance = dolphie.metric_manager.metrics.__dict__.get(metric_name)
                            if metric_instance:
                                for field_name, metric_values in metric_data.items():
                                    metric: MetricManager.MetricData = metric_instance.__dict__.get(field_name)
                                    if metric and metric_values:
                                        metric.values.extend(metric_values)
                                        metric.last_value = metric_values[-1]
            else:
                # Full format: replace values entirely
                dolphie.metric_manager.datetimes = deque(new_datetimes)
                for metric_name, metric_data in replay_event_data.metric_manager.items():
                    metric_instance = dolphie.metric_manager.metrics.__dict__.get(metric_name)
                    if metric_instance:
                        for field_name, metric_values in metric_data.items():
                            metric: MetricManager.MetricData = metric_instance.__dict__.get(field_name)
                            if metric:
                                metric.values = deque(metric_values)
                                metric.last_value = metric_values[-1]

        except Exception as e:
            # Catch any errors during replay and log them without crashing the app
            self.app.notify(
                f"Error during replay: {str(e)}",
                title="Replay Error",
                severity="error",
            )
            if tab.worker:
                tab.worker.cancel()

    async def run_worker_main(self, tab_id: str):
        tab = self.app.tab_manager.get_tab(tab_id)
        if not tab:
            return

        # Get our worker thread
        tab.worker = get_current_worker()
        tab.worker.name = tab_id

        dolphie = tab.dolphie
        try:
            if not dolphie.main_db_connection.is_connected():
                # Update connection status from worker thread
                self.app.call_from_thread(
                    self.app.tab_manager.update_connection_status,
                    tab=tab,
                    connection_status=ConnectionStatus.connecting
                )

                tab.replay_manager = None
                if not dolphie.daemon_mode and tab == self.app.tab_manager.active_tab:
                    # Display property triggers UI updates, must be called from main thread
                    def show_loading():
                        tab.loading_indicator.display = True
                    self.app.call_from_thread(show_loading)

                dolphie.db_connect()

            worker_start_time = datetime.now().astimezone()
            dolphie.polling_latency = (worker_start_time - dolphie.worker_previous_start_time).total_seconds()
            dolphie.worker_previous_start_time = worker_start_time

            dolphie.collect_system_utilization()
            if dolphie.connection_source == ConnectionSource.mysql:
                self.app.worker_data_processor.process_mysql_data(tab)
            elif dolphie.connection_source == ConnectionSource.proxysql:
                self.app.worker_data_processor.process_proxysql_data(tab)

            dolphie.worker_processing_time = (datetime.now().astimezone() - worker_start_time).total_seconds()

            dolphie.metric_manager.refresh_data(
                worker_start_time=worker_start_time,
                polling_latency=dolphie.polling_latency,
                system_utilization=dolphie.system_utilization,
                global_variables=dolphie.global_variables,
                global_status=dolphie.global_status,
                innodb_metrics=dolphie.innodb_metrics,
                disk_io_metrics=dolphie.disk_io_metrics,
                metadata_lock_metrics=dolphie.metadata_locks,
                replication_status=dolphie.replication_status,
                proxysql_command_stats=dolphie.proxysql_command_stats,
            )

            # We initalize this here so we have the host version from process_{mysql,proxysql}_data
            if not tab.replay_manager:
                tab.replay_manager = ReplayManager(dolphie)

            tab.replay_manager.capture_state()
        except ManualException as exception:
            # This will set up the worker state change function below to trigger the
            # tab setup modal with the error
            tab.worker_cancel_error = exception

            # Disconnect from worker thread - call_from_thread handles async functions
            self.app.call_from_thread(self.app.tab_manager.disconnect_tab, tab)

    def run_worker_replicas(self, tab_id: str):
        tab = self.app.tab_manager.get_tab(tab_id)
        if not tab:
            return

        # Get our worker thread
        tab.replicas_worker = get_current_worker()
        tab.replicas_worker.name = tab_id

        dolphie = tab.dolphie

        if dolphie.panels.replication.visible:
            if tab.id != self.app.tab_manager.active_tab.id:
                return

            if dolphie.replica_manager.available_replicas:
                if not dolphie.replica_manager.replicas:
                    def update_replicas_ui():
                        tab.replicas_container.display = True
                        tab.replicas_loading_indicator.display = True
                        tab.replicas_title.update(
                            f"[$white][b]Loading [$highlight]{len(dolphie.replica_manager.available_replicas)}"
                            "[/$highlight] replicas...\n"
                        )
                    self.app.call_from_thread(update_replicas_ui)

                ReplicationPanel.fetch_replicas(tab)
            else:
                # Display property triggers UI updates, must be called from main thread
                def hide_replicas():
                    tab.replicas_container.display = False
                self.app.call_from_thread(hide_replicas)
        else:
            # If we're not displaying the replication panel, remove all replica connections
            dolphie.replica_manager.remove_all_replicas()

    def on_worker_state_changed(self, event: Worker.StateChanged):
        if event.state not in [WorkerState.SUCCESS, WorkerState.CANCELLED]:
            return

        tab = self.app.tab_manager.get_tab(event.worker.name)
        if not tab:
            return

        dolphie = tab.dolphie

        if event.worker.group == "main":
            if event.state == WorkerState.SUCCESS:
                self.app.worker_data_processor.monitor_read_only_change(tab)

                refresh_interval = dolphie.refresh_interval
                if dolphie.connection_source == ConnectionSource.proxysql:
                    refresh_interval = dolphie.determine_proxysql_refresh_interval()

                # Skip this if the conditions are right
                if (
                    len(self.app.screen_stack) > 1
                    or dolphie.pause_refresh
                    or not dolphie.main_db_connection.is_connected()
                    or dolphie.daemon_mode
                    or tab.id != self.app.tab_manager.active_tab.id
                ):
                    tab.worker_timer = self.app.set_timer(refresh_interval, partial(self.app.run_worker_main, tab.id))

                    return

                if not tab.main_container.display:
                    tab.toggle_metric_graph_tabs_display()

                if dolphie.connection_source == ConnectionSource.mysql:
                    self.app.worker_data_processor.refresh_screen_mysql(tab)
                elif dolphie.connection_source == ConnectionSource.proxysql:
                    self.app.worker_data_processor.refresh_screen_proxysql(tab)

                # Update the topbar with the latest replay file size
                if dolphie.record_for_replay:
                    self.app.tab_manager.update_topbar(tab=tab)

                tab.toggle_entities_displays()

                tab.worker_timer = self.app.set_timer(refresh_interval, partial(self.app.run_worker_main, tab.id))
            elif event.state == WorkerState.CANCELLED:
                # Only show the modal if there's a worker cancel error
                if tab.worker_cancel_error:
                    from loguru import logger

                    logger.critical(tab.worker_cancel_error)

                    if self.app.tab_manager.active_tab.id != tab.id or self.app.tab_manager.loading_hostgroups:
                        self.app.notify(
                            (
                                f"[$b_light_blue]{dolphie.host}:{dolphie.port}[/$b_light_blue]: "
                                f"{tab.worker_cancel_error.reason}"
                            ),
                            title="Connection Error",
                            severity="error",
                            timeout=10,
                        )

                    if not self.app.tab_manager.loading_hostgroups:
                        self.app.tab_manager.switch_tab(tab.id)

                        self.app.tab_manager.setup_host_tab(tab)
                        self.app.bell()
        elif event.worker.group == "replicas":
            if event.state == WorkerState.SUCCESS:
                # Skip this if the conditions are right
                if (
                    len(self.app.screen_stack) > 1
                    or dolphie.pause_refresh
                    or tab.id != self.app.tab_manager.active_tab.id
                ):
                    tab.replicas_worker_timer = self.app.set_timer(
                        dolphie.refresh_interval,
                        partial(self.app.run_worker_replicas, tab.id),
                    )
                    return

                if dolphie.panels.replication.visible and dolphie.replica_manager.available_replicas:
                    ReplicationPanel.create_replica_panel(tab)

                tab.replicas_worker_timer = self.app.set_timer(
                    dolphie.refresh_interval,
                    partial(self.app.run_worker_replicas, tab.id),
                )
        elif event.worker.group == "replay" and event.state == WorkerState.SUCCESS:
            if tab.id == self.app.tab_manager.active_tab.id:
                if len(self.app.screen_stack) > 1 or (dolphie.pause_refresh and not tab.replay_manual_control):
                    tab.worker_timer = self.app.set_timer(
                        dolphie.refresh_interval,
                        partial(self.app.run_worker_replay, tab.id),
                    )

                    return
            else:
                # If the tab isn't active, stop the loop
                return

            self.app.worker_data_processor.monitor_read_only_change(tab)

            if not tab.main_container.display:
                tab.toggle_metric_graph_tabs_display()

            if dolphie.connection_source == ConnectionSource.mysql:
                self.app.worker_data_processor.refresh_screen_mysql(tab)
                ReplicationPanel.create_replica_panel(tab)
            elif dolphie.connection_source == ConnectionSource.proxysql:
                self.app.worker_data_processor.refresh_screen_proxysql(tab)

            tab.toggle_entities_displays()

            tab.worker_timer = self.app.set_timer(
                dolphie.refresh_interval,
                partial(self.app.run_worker_replay, tab.id),
            )



from __future__ import annotations

from dataclasses import dataclass, field

import pymysql
from dolphie.Modules.Functions import format_query, format_time
from rich.table import Table


class ConnectionSource:
    mysql = "MySQL"
    proxysql = "ProxySQL"
    mariadb = "MariaDB"


class ConnectionStatus:
    connecting = "CONNECTING"
    connected = "CONNECTED"
    disconnected = "DISCONNECTED"
    read_write = "R/W"
    read_only = "RO"


@dataclass
class Replica:
    row_key: str
    host: str
    thread_id: int | None = None
    port: int | None = None
    host_distro: str | None = None
    connection: pymysql.Connection | None = None
    connection_source_alt: ConnectionSource | None = None
    table: Table | None = None
    replication_status: dict[str, str | int] = field(default_factory=dict)
    mysql_version: str | None = None


class ReplicaManager:
    def __init__(self):
        self.available_replicas: list[dict[str, str]] = []
        self.replicas: dict[str, Replica] = {}
        self.ports: dict[str, dict[str, str | int | bool]] = {}

    # Dots/colons are invalid in Textual widget IDs - translate to hyphens in one pass
    _widget_id_sanitize = str.maketrans({".": "-", ":": "-"})

    @classmethod
    def create_replica_row_key(cls, host: str, port: int) -> str:
        return f"{host}-{port}".translate(cls._widget_id_sanitize)

    def add_replica(self, row_key: str, thread_id: int, host: str, port: int) -> Replica:
        self.replicas[row_key] = Replica(row_key=row_key, thread_id=thread_id, host=host, port=port)
        return self.replicas[row_key]

    def remove_replica(self, row_key: str):
        replica = self.replicas.pop(row_key, None)
        if replica and replica.connection:
            replica.connection.close()

    def get_replica(self, row_key: str) -> Replica | None:
        return self.replicas.get(row_key)

    def remove_all_replicas(self):
        for replica in self.replicas.values():
            if replica.connection:
                replica.connection.close()
        self.replicas = {}

    def get_sorted_replicas(self) -> list[Replica]:
        return sorted(self.replicas.values(), key=lambda x: x.host)


@dataclass
class Panel:
    name: str
    display_name: str
    key: str | None = None
    visible: bool = False
    daemon_supported: bool = True

    @property
    def formatted_key(self) -> str:
        # Rich markup — used in Rich Table titles (Dashboard, ProxySQL)
        return f"[b highlight]{self.key}[/b highlight]"

    @property
    def content_key(self) -> str:
        # Textual Content markup — used in .update() calls
        return f"[$b_highlight]{self.key}[/$b_highlight]"

    @property
    def title(self) -> str:
        return f"[$b_highlight]{self.key}[/$b_highlight]{self.display_name}"


class Panels:
    def __init__(self):
        self._registry: dict[str, Panel] = {}

        self.dashboard = self._add("dashboard", "Dashboard", "¹", daemon_supported=False)
        self.processlist = self._add("processlist", "Processlist", "²")
        self.graphs = self._add("graphs", "Metric Graphs", "³", daemon_supported=False)
        self.replication = self._add("replication", "Replication", "⁴", daemon_supported=False)
        self.metadata_locks = self._add("metadata_locks", "Metadata Locks", "⁵")
        self.ddl = self._add("ddl", "DDL", "⁶", daemon_supported=False)
        self.pfs_metrics = self._add("pfs_metrics", "Performance Schema Metrics", "⁷")
        self.statements_summary = self._add("statements_summary", "Statements Summary", "⁸")
        self.proxysql_hostgroup_summary = self._add("proxysql_hostgroup_summary", "Hostgroup Summary", "⁴")
        self.proxysql_mysql_query_rules = self._add(
            "proxysql_mysql_query_rules", "Query Rules", "⁵", daemon_supported=False
        )
        self.proxysql_command_stats = self._add("proxysql_command_stats", "Command Stats", "⁶", daemon_supported=False)

    def _add(self, name: str, display_name: str, key: str | None = None, daemon_supported: bool = True) -> Panel:
        panel = Panel(name=name, display_name=display_name, key=key, daemon_supported=daemon_supported)
        self._registry[name] = panel
        return panel

    def validate_panels(self, panel_list_str: str | list[str], valid_panel_names: list[str]) -> list[str]:
        panels = panel_list_str.split(",") if isinstance(panel_list_str, str) else panel_list_str

        invalid_panels = [panel for panel in panels if panel not in valid_panel_names]
        if invalid_panels:
            raise ValueError(
                f"Panel(s) [red2]{', '.join(invalid_panels)}[/red2] are not valid (see --help for more information)"
            )

        return panels

    def get_all_daemon_panel_names(self) -> list[str]:
        return [panel.name for panel in self._registry.values() if panel.daemon_supported]

    def get_all_panels(self) -> list[Panel]:
        return list(self._registry.values())

    def all(self) -> list[str]:
        return list(self._registry.keys())


class BaseProcesslistThread:
    def __init__(self, thread_data: dict[str, str]):
        self.thread_data = thread_data
        self.id = str(thread_data.get("id", ""))
        self.user = thread_data.get("user", "")
        self.db = thread_data.get("db", "")

    @staticmethod
    def _get_time_color(time: float, query_code: str) -> str:
        if not query_code:
            return ""
        if time >= 10:
            return "red"
        elif time >= 5:
            return "yellow"
        return "green"

    @staticmethod
    def _format_time_with_color(time: float, color: str) -> str:
        return f"[{color}]{format_time(time)}[/{color}]" if color else format_time(time)

    @staticmethod
    def _format_command(command: str) -> str:
        return "[red]Killed[/red]" if command == "Killed" else command

    @staticmethod
    def _format_string(string: str) -> str:
        return string if string else "[dark_gray]N/A"

    @staticmethod
    def _format_number(number) -> str:
        return "[dark_gray]0" if not number or number == "0" else number


class ProcesslistThread(BaseProcesslistThread):
    def __init__(self, thread_data: dict[str, str]):
        super().__init__(thread_data)

        self.mysql_thread_id = thread_data.get("mysql_thread_id")
        self.host = thread_data.get("host", "")
        self.time = int(thread_data.get("time", 0))
        self.protocol = self._format_string(thread_data.get("connection_type", ""))
        self.formatted_query = format_query(thread_data.get("query", ""))
        self.formatted_time = self._format_time_with_color(self.time, self._mysql_time_color())
        self.command = self._format_command(thread_data.get("command", ""))
        self.state = self._format_string(thread_data.get("state", ""))
        self.trx_state = self._format_string(thread_data.get("trx_state", ""))
        self.trx_operation_state = self._format_string(thread_data.get("trx_operation_state", ""))
        self.trx_rows_locked = self._format_number(thread_data.get("trx_rows_locked", 0))
        self.trx_rows_modified = self._format_number(thread_data.get("trx_rows_modified", 0))
        self.trx_concurrency_tickets = self._format_number(thread_data.get("trx_concurrency_tickets", 0))
        trx_time = thread_data.get("trx_time", "")
        self.trx_time = format_time(int(trx_time)) if trx_time else "[dark_gray]N/A"

    def _mysql_time_color(self) -> str:
        if "Group replication" in self.formatted_query.code:  # Don't color GR threads
            return ""
        if "SELECT /*!40001 SQL_NO_CACHE */ *" in self.formatted_query.code:
            return "purple"
        return self._get_time_color(self.time, self.formatted_query.code)


class ProxySQLProcesslistThread(BaseProcesslistThread):
    def __init__(self, thread_data: dict[str, str]):
        super().__init__(thread_data)

        self.hostgroup = int(thread_data.get("hostgroup"))
        self.frontend_host = self._format_string(thread_data.get("frontend_host", ""))
        self.host = self._format_string(thread_data.get("backend_host", ""))
        self.time = int(thread_data.get("time", 0)) / 1000  # Convert to seconds since ProxySQL returns milliseconds
        self.formatted_query = format_query(thread_data.get("query", "").strip(" \t\n\r"))
        color = self._get_time_color(self.time, self.formatted_query.code)
        self.formatted_time = self._format_time_with_color(self.time, color)
        self.command = self._format_command(thread_data.get("command", ""))
        self.extended_info = thread_data.get("extended_info", "")


class HotkeyCommands:
    show_thread = "show_thread"
    thread_filter = "thread_filter"
    thread_kill_by_parameter = "thread_kill_by_parameter"
    variable_search = "variable_search"
    rename_tab = "rename_tab"
    refresh_interval = "refresh_interval"
    replay_seek = "replay_seek"
    maximize_panel = "maximize_panel"
